# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, Gio


class Action(Gio.SimpleAction):
    def __init__(self, name, activate_cb=None, restrict=False, shared=None, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        if activate_cb:
            self.connect('activate', activate_cb)
        if restrict:
            shared.bind_property('restricted', self, 'enabled', GObject.BindingFlags.SYNC_CREATE, lambda x, y: not y)


class ActionModel(object):
    def __init__(self, name, activate_cb, *args, **kwargs):
        self.name = name
        self.activate_cb = activate_cb
        self.args = args
        self.kwargs = kwargs

    def create_action(self, decorator, **kwargs):
        return Action(self.name, decorator(self.activate_cb), *self.args, **self.kwargs, **kwargs)

    def get_name(self):
        return self.name


class PropertyAction(Gio.PropertyAction):
    def __init__(self, name, obj, property_name=None, **kwargs):
        super().__init__(name=name, object=obj, property_name=property_name or name, **kwargs)



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GLib

import ampd

from gampc import plugin
from gampc.utils import action
from gampc.utils import omenu


class CoreExtension(plugin.Extension):
    def activate(self):
        self.ampd = self.shared.ampd
        self.fading = False
        self.provides['app'] = {}
        self.provides['app']['actions'] = [
            action.PropertyAction('restricted', self.shared),
            action.PropertyAction('dark', self.shared),
            action.PropertyAction('server-profile', self.shared),
            action.ActionModel('connect', self.shared.ampd_connect),
            action.ActionModel('disconnect', self.shared.ampd_disconnect),
            action.ActionModel('play-or-pause', self.play_or_pause_cb, restrict=True),
            action.ActionModel('absolute-jump', self.absolute_jump_cb, restrict=True, parameter_type=GLib.VariantType.new('i')),
            action.ActionModel('relative-jump', self.relative_jump_cb, restrict=True, parameter_type=GLib.VariantType.new('i')),
        ] + [
            action.ActionModel(name, self.mpd_command_cb, restrict=True) for name in
            ('play', 'stop', 'next', 'previous')
        ] + [
            action.ActionModel(name, self.mpd_command_cb) for name in
            ('update',)
        ] + [action.ActionModel('fade-to-' + name, self.fade_to_action_cb) for name in ('next', 'stop')] + [action.PropertyAction('option-' + option, self.shared.ampd_server_properties) for option in self.shared.ampd_server_properties.OPTION_NAMES]

        self.provides['app']['menubar_items'] = [
            omenu.Item('10#gampc/20#misc/50', 'app.restricted', _("Restricted mode"), ['<Control><Alt><Shift>s']),
            omenu.Item('10#gampc/20#misc/60', 'app.dark', _("Dark interface"), ['<Control><Alt><Shift>d']),
            omenu.Item('30#playback/10#play/10', 'app.play-or-pause', _("_Play/pause"), ['<Control>Up', 'AudioPlay', 'space'], accels_fragile=True),
            omenu.Item('30#playback/10#play/20', 'app.stop', _("_Stop"), ['<Control>Down', 'AudioStop'], accels_fragile=True),
            omenu.Item('30#playback/10#play/30', 'app.fade-to-stop', _("Fade to stop"), ['<Control><Shift>Down', '<Shift>AudioStop'], accels_fragile=True),
            omenu.Item('30#playback/20#move/10', 'app.previous', _("_Previous"), ['<Control>Left', 'AudioPrev'], accels_fragile=True),
            omenu.Item('30#playback/20#move/20', 'app.next', _("_Next"), ['<Control>Right', 'AudioNext'], accels_fragile=True),
            omenu.Item('30#playback/20#move/30', 'app.fade-to-next', _("_Fade to next"), ['<Control><Shift>Right'], accels_fragile=True),
            omenu.Item('30#playback/30#jump/10', 'app.absolute-jump(0)', _("Restart playback"), ['<Alt>Up']),
            omenu.Item('30#playback/30#jump/20', 'app.absolute-jump(-15)', _("End of song (-{} seconds)").format(15), ['<Alt>Down']),
            omenu.Item('30#playback/30#jump/30', 'app.relative-jump(-5)', _("Skip backwards ({} seconds)").format(5), ['<Alt>Left'], accels_fragile=True),
            omenu.Item('30#playback/30#jump/40', 'app.relative-jump(5)', _("Skip forwards ({} seconds)").format(5), ['<Alt>Right'], accels_fragile=True),
            omenu.Item('40#server/10#database/10', 'app.update', _("Update database")),
            omenu.Item('40#server/20#options/10', 'app.option-random', _("Random mode")),
            omenu.Item('40#server/20#options/20', 'app.option-repeat', _("Repeat mode")),
            omenu.Item('40#server/20#options/30', 'app.option-consume', _("Consume mode")),
            omenu.Item('40#server/20#options/40', 'app.option-single', _("Single mode")),
            omenu.Item('40#server/40#connection/10', 'app.connect', _("Connect"), ['<Alt><Shift>c']),
            omenu.Item('40#server/40#connection/20', 'app.disconnect', _("Disconnect"), ['<Alt><Shift>d']),
        ]

    @ampd.task
    async def mpd_command_cb(self, caller, *data):
        if not self.shared.ampd_server_properties.state:
            await self.ampd.idle(ampd.IDLE)
        await getattr(self.ampd, caller.get_name())()

    @ampd.task
    async def play_or_pause_cb(self, action_, parameter):
        if not self.shared.ampd_server_properties.state:
            await self.ampd.idle(ampd.IDLE)
        await (self.ampd.pause(1) if self.shared.ampd_server_properties.state == 'play' else self.ampd.play())

    @ampd.task
    async def fade_to_action_cb(self, action_, parameter):
        if self.fading:
            return
        try:
            self.fading = True
            N = 30
            T = 5

            if not self.shared.ampd_server_properties.state:
                await self.ampd.idle(ampd.IDLE)
            if self.shared.ampd_server_properties.state != 'play':
                return
            volume = self.shared.ampd_server_properties.volume
            for i in range(N):
                self.shared.ampd_server_properties.volume = volume * (N - i - 1) / N
                reply = await self.ampd.idle(ampd.PLAYER, timeout=T / N)
                if reply & ampd.PLAYER:
                    break
            else:
                await getattr(self.ampd, action_.get_name()[8:])()
            self.shared.ampd_server_properties.volume = volume
        finally:
            self.fading = False

    def absolute_jump_cb(self, action_, parameter):
        target = parameter.unpack()
        if self.shared.ampd_server_properties.state != 'stop':
            self.shared.ampd_server_properties.elapsed = target if target >= 0 else self.shared.ampd_server_properties.time + target

    def relative_jump_cb(self, action_, parameter):
        if self.shared.ampd_server_properties.state != 'stop':
            self.shared.ampd_server_properties.elapsed += parameter.unpack()



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GLib
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Gio
import asyncio
import sys

import ampd

from gampc import data
from gampc import module
from gampc import plugin
from gampc.utils import action
from gampc.utils import omenu
from gampc.utils import config
from gampc.utils.logger import logger


class SongList(module.Module):
    sortable = False
    duplicate_test_columns = []
    duplicate_field = '_duplicate'
    use_provided = ['songlist']

    CAPABILITIES_DRAG_DEST = 1
    CAPABILITIES_DRAG_SOURCE = 2

    capabilities = 0

    def __init__(self, shared):
        super().__init__(shared)
        self.fields = self.shared.songlist_fields

        self.treeview = data.RecordTreeView(self.fields, self.data_func, self.sortable)

        self.store = self.treeview.get_model()

        self.actions.add_action(action.Action('reset', self.action_reset_cb))
        self.actions.add_action(action.Action('copy', self.action_copy_delete_cb))

        dndtargets = [Gtk.TargetEntry.new('GAMPC_SONGS', Gtk.TargetFlags(0), 0)]

        if self.new_song != NotImplemented:
            self.actions.add_action(action.Action('paste', self.action_paste_cb))
            self.actions.add_action(action.Action('paste-before', self.action_paste_cb))
            self.treeview.drag_dest_set(Gtk.DestDefaults.DROP, dndtargets, Gdk.DragAction.MOVE | Gdk.DragAction.COPY)
            self.signal_handler_connect(self.store, 'record-new', self.new_song)

        if self.delete_song != NotImplemented:
            self.actions.add_action(action.Action('delete', self.action_copy_delete_cb))
            self.actions.add_action(action.Action('cut', self.action_copy_delete_cb))
            self.treeview.drag_source_set(Gdk.ModifierType.BUTTON1_MASK, dndtargets, Gdk.DragAction.MOVE | Gdk.DragAction.COPY)
            self.signal_handler_connect(self.store, 'record-delete', self.delete_song)
        else:
            self.treeview.drag_source_set(Gdk.ModifierType.BUTTON1_MASK, dndtargets, Gdk.DragAction.COPY)

        self.treeview_filter = data.TreeViewFilter(shared, self.treeview)
        self.add(self.treeview_filter)
        self.actions.add_action(action.PropertyAction('filter', self.treeview_filter, 'active'))

        self.setup_context_menu(self.provided.get('context_menu_items', []), self.treeview)
        self.treeview.connect('row-activated', self.treeview_row_activated_cb)

        self.connect('map', self.set_color)
        self.signal_handler_connect(self.shared, 'notify::dark', self.set_color)

    def set_color(self, *args):
        if self.win:
            self.color = self.win.get_style_context().get_color(Gtk.StateFlags.NORMAL)

    def action_copy_delete_cb(self, action, parameter):
        songs, refs = self.treeview.get_selection_rows()
        if action.get_name() in ['copy', 'cut']:
            Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).set_text(repr(songs), -1)
        if action.get_name() in ['delete', 'cut']:
            self.store.delete_refs(refs)

    def action_paste_cb(self, action, parameter):
        Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).request_text(self.treeview.clipboard_paste_cb, action.get_name().endswith('before'))

    @ampd.task
    async def treeview_row_activated_cb(self, treeview, p, column):
        if self.shared.restricted:
            return
        filename = self.store.get_record(self.store.get_iter(p)).file
        songs = await self.ampd.playlistfind('file', filename)
        if songs:
            song_id = sorted(songs, key=lambda song: song['Pos'])[0]['Id']
        else:
            song_id = await self.ampd.addid(filename)
        await self.ampd.playid(song_id)

    def get_filenames(self, selection=True):
        if selection:
            store, paths = self.treeview.get_selection().get_selected_rows()
            rows = (store.get_record(store.get_iter(p)) for p in paths)
        else:
            rows = (self.store.get_record(self.store.iter_nth_child(None, i)) for i in range(self.store.iter_n_children()))
        return (row.file for row in rows if row)

    def _mix_colors(self, r, g, b):
        return Gdk.RGBA((1 - self.color.red) * r + 0.5 * (1 - r),
                        (1 - self.color.green) * g + 0.5 * (1 - g),
                        (1 - self.color.blue) * b + 0.5 * (1 - b),
                        1.0)

    def _mix_colors_rgba(self, rgba):
        return self._mix_colors(rgba.red, rgba.green, rgba.blue)

    def data_func(self, column, renderer, store, i, j):
        _duplicate = store.get_record(i)._data.get(self.duplicate_field, None)
        if _duplicate is None:
            renderer.set_property('background-rgba', None)
        else:
            N = 4
            m = _duplicate % (N ** 3 - 1)
            renderer.set_property('background-rgba',
                                  self._mix_colors(*(float((m // (N ** k)) % N) / (N - 1)
                                                     for k in (2, 1, 0))))

    def set_songs(self, songs, set_fields=True):
        if set_fields:
            self.records_set_fields(songs)
        if self.duplicate_test_columns:
            self.find_duplicates(songs, self.duplicate_test_columns)
        self.store.set_rows(songs)

    def records_set_fields(self, songs):
        self.fields.records_set_fields(songs)

    def find_duplicates(self, songs, test_fields):
        dup_marker = 0
        dup_dict = {}
        for song in songs:
            if 'Time' not in song or int(song['Time']) <= 3:
                continue
            test = tuple(song.get(field) for field in test_fields)
            duplicates = dup_dict.get(test)
            if duplicates:
                if len(duplicates) == 1:
                    duplicates[0][self.duplicate_field] = dup_marker
                    dup_marker += 1
                song[self.duplicate_field] = duplicates[0][self.duplicate_field]
                duplicates.append(song)
            else:
                dup_dict[test] = [song]
                song.pop(self.duplicate_field, None)

    def action_reset_cb(self, action, parameter):
        self.treeview_filter.filter_.set_data({})
        self.treeview_filter.active = False
        if self.sortable:
            self.store.set_sort_column_id(-1, Gtk.SortType.ASCENDING)

    delete_song = new_song = NotImplemented


class SongListWithTotals(SongList):
    def set_songs(self, songs, set_fields=True):
        super().set_songs(songs, set_fields)
        time = sum(int(song.get('Time', '0')) for song in songs)
        self.status = '{} / {}'.format(len(songs), data.format_time(time))


class EditableSongListBase(SongList):
    SONG_NEW = 1
    SONG_DELETED = 2
    SONG_MODIFIED = 3
    SONG_UNDEFINED = 4

    STATUS_PROPERTIES = ('background-rgba', 'font', 'strikethrough')
    STATUS_PROPERTY_TABLE = {
        SONG_NEW: (Gdk.RGBA(0.0, 1.0, 0.0, 1.0), 'bold', None),
        SONG_DELETED: (Gdk.RGBA(1.0, 0.0, 0.0, 1.0), 'italic', True),
        SONG_MODIFIED: (Gdk.RGBA(1.0, 1.0, 0.0, 1.0), None, None),
        SONG_UNDEFINED: (None, 'bold italic', None),
    }

    def __init__(self, shared):
        super().__init__(shared)
        self.actions.add_action(action.Action('save', self.action_save_cb))
        self.actions.add_action(action.Action('undelete', self.action_undelete_cb))

    def data_func(self, column, renderer, store, i, j):
        for p in self.STATUS_PROPERTIES:
            renderer.set_property(p, None)
        super().data_func(column, renderer, store, i, j)
        status = store.get_record(i)._status
        if status is not None:
            for k, p in enumerate(self.STATUS_PROPERTIES):
                if self.STATUS_PROPERTY_TABLE[status][k] is not None:
                    renderer.set_property(p, self.STATUS_PROPERTY_TABLE[status][k])

    def action_undelete_cb(self, action, parameter):
        store, paths = self.treeview.get_selection().get_selected_rows()
        for p in paths:
            i = self.store.get_iter(p)
            if self.store.get_record(i)._status == self.SONG_DELETED:
                del self.store.get_record(i)._status
        self.treeview.queue_draw()

    def delete_song(self, store, i):
        if store.get_record(i)._status == self.SONG_UNDEFINED:
            return
        self.set_modified()
        if store.get_record(i)._status == self.SONG_NEW:
            store.remove(i)
        else:
            self.store.get_record(i)._status = self.SONG_DELETED
            self.merge_new_del(i)
        self.treeview.queue_draw()

    def merge_new_del(self, i):
        _status = self.store.get_record(i)._status
        for f in [self.store.iter_previous, self.store.iter_next]:
            j = f(i)
            if j and self.store.get_record(j).file == self.store.get_record(i).file and {_status, self.store.get_record(j)._status} == {self.SONG_DELETED, self.SONG_NEW}:
                del self.store.get_record(i)._status
                self.store.remove(j)
                return


class EditableSongList(EditableSongListBase):
    def new_song(self, store, i):
        self.set_modified()
        store.get_record(i)._status = self.SONG_NEW
        self.merge_new_del(i)


class EditableFileSongList(EditableSongListBase):
    def records_set_fields(self, songs):
        for song in songs:
            gfile = Gio.File.new_for_path(GLib.build_filenamev([self.shared.config.songlist.music_dir, song['file']]))
            if gfile.query_exists():
                song['_gfile'] = gfile
            else:
                song['_status'] = self.SONG_UNDEFINED
        super().records_set_fields(songs)

    def action_save_cb(self, action, parameter):
        self.save_files(song for i, p, song in self.store)

    def action_save_selected_cb(self, action, parameter):
        songs, refs = self.treeview.get_selection_rows()

    def save_files(self, songs):
        deleted = [song for song in songs if song._status == self.SONG_DELETED]
        if deleted:
            dialog = Gtk.Dialog(parent=self.win, title=_("Move to trash"))
            dialog.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
            dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
            dialog.get_content_area().add(Gtk.Label(label='\n\t'.join([_("Move these files to the trash?")] + [song.file for song in deleted]), visible=True))
            reply = dialog.run()
            dialog.destroy()
            if reply != Gtk.ResponseType.OK:
                return
            for song in deleted:
                song._gfile.trash()
                song._status = self.SONG_UNDEFINED

    def set_modified(self):
        self.status = _("modified")

    def set_songs(self, songs, set_fields=True):
        self.status = None
        super().set_songs(songs, set_fields)


class SongListExtension(plugin.Extension):
    def __init__(self):
        super().__init__()
        self.provides['app'] = {}
        self.provides['app']['menubar_items'] = [
            omenu.Item('20#edit/10#global/10', 'mod.save', _("Save"), ['<Control>s']),
            omenu.Item('20#edit/10#global/20', 'mod.reset', _("Reset"), ['<Control>r']),
            omenu.Item('20#edit/10#global/30', 'mod.filter', _("Filter"), ['<Control><Shift>f']),
        ]
        self.provides['songlist'] = {}
        self.provides['songlist']['context_menu_items'] = []

        for items, prefix in ((self.provides['app']['menubar_items'], '20#edit/20#local/'),
                              (self.provides['songlist']['context_menu_items'], '50#edit/')):
            items += [
                omenu.Item(prefix + '10', 'mod.cut', _("Cut"), ['<Control>x'], accels_fragile=True),
                omenu.Item(prefix + '20', 'mod.copy', _("Copy"), ['<Control>c'], accels_fragile=True),
                omenu.Item(prefix + '30', 'mod.paste', _("Paste"), ['<Control>v'], accels_fragile=True),
                omenu.Item(prefix + '40', 'mod.paste-before', _("Paste before"), ['<Control>b']),
                omenu.Item(prefix + '50', 'mod.delete', _("Delete"), ['Delete'], accels_fragile=True),
                omenu.Item(prefix + '60', 'mod.undelete', _("Undelete"), ['<Alt>Delete'], accels_fragile=True),
            ]

    # try:
    #     import mutagen
    # except:
    #     mutagen = None

    @staticmethod
    def song_title(song):
        title = song.get('Title') or song.get('Name', '')
        filename = song.get('file', '')
        if filename.startswith('http://'):
            http_title = filename[filename.rfind('/') + 1:]
            http_title = http_title[:http_title.rfind('.')]
            title = '{0} [{1}]'.format(title, http_title) if title else http_title
        return title

    def get_mutagen_file(self, song):
        return None
        if self.mutagen is None:
            return None
        try:
            return self.mutagen.File(song['_gfile'].get_path())
        except:
            return None

    @staticmethod
    def get_mutagen_bitrate(song):
        if '_mutagen' not in song:
            return None
        try:
            return str(song['_mutagen'].info.bitrate // 1000)
        except:
            if song['file'].endswith('.flac'):
                return 'FLAC'
            else:
                return '???'

    def activate(self):
        self.config = config.ConfigDict.load('songlist', self.shared.config)
        self.config.access('music-dir', GLib.get_user_special_dir(GLib.USER_DIRECTORY_MUSIC))

        self.shared.songlist_fields = data.FieldFamily(self.config.fields)
        self.shared.songlist_fields.register_field(data.Field('Album', _("Album")))
        self.shared.songlist_fields.register_field(data.Field('AlbumArtist', _("Album artist")))
        self.shared.songlist_fields.register_field(data.Field('Artist', _("Artist")))
        self.shared.songlist_fields.register_field(data.Field('Composer', _("Composer")))
        self.shared.songlist_fields.register_field(data.Field('Date', _("Date")))
        self.shared.songlist_fields.register_field(data.Field('Disc', _("Disc")))
        self.shared.songlist_fields.register_field(data.Field('file', _("File")))
        self.shared.songlist_fields.register_field(data.Field('Genre', _("Genre")))
        self.shared.songlist_fields.register_field(data.Field('Performer', _("Performer")))
        self.shared.songlist_fields.register_field(data.Field('Time', _("Seconds"), visible=False))
        self.shared.songlist_fields.register_field(data.Field('FormattedTime', _("Duration"), get_value=lambda song: data.format_time(song['Time']) if 'Time' in song else ''))
        self.shared.songlist_fields.register_field(data.Field('Title', _("Title (partial)")))
        self.shared.songlist_fields.register_field(data.Field('FullTitle', _("Title"), get_value=self.song_title))
        self.shared.songlist_fields.register_field(data.Field('Track', _("Track")))
        # self.shared.songlist_fields.register_field(data.Field('_mutagen', get_value=self.get_mutagen_file, visible=False))
        # self.shared.songlist_fields.register_field(data.Field('_mutagen_bitrate', _('Bitrate'), get_value=self.get_mutagen_bitrate, xalign=1.0))
        self.shared.songlist_fields.register_field(data.FieldWithTable(
            'Extension', _("Extension"),
            table=[
                [
                    'file',
                    '\\.([^.]*)$',
                    '\\1'
                ]
            ]))
        self.shared.songlist_fields.register_field(data.FieldWithTable(
            'agenre', visible=False,
            table=[
                [
                    'Genre',
                    '[Mm]ilong',
                    'b milonga'
                ],
                [
                    'Genre',
                    '[Cc]andombe',
                    'b milonga'
                ],
                [
                    'Genre',
                    '[Tt]ango|Canci[oó]n',
                    'a tango'
                ],
                [
                    'Genre',
                    '[Vv]als',
                    'c vals'
                ],
                [
                    'Genre',
                    '[Ff]ox ?trot',
                    'd fox'
                ],
                [
                    'Genre',
                    '[Pp]aso ?doble',
                    'e paso'
                ],
                [
                    'Genre',
                    'Ranchera',
                    'f ranchera'
                ],
                [
                    None,
                    None,
                    'z'
                ]
            ]))
        self.shared.songlist_fields.register_field(data.FieldWithTable(
            'ArtistSortName', visible=False,
            table=[
                [
                    'Artist',
                    '(La Típica Sanata|Otros Aires|.* Orquesta)',
                    '\\1'
                ],
                [
                    'Artist',
                    '^(.* Tango)$',
                    '\\1'
                ],
                [
                    'Artist',
                    '(.*), dir\. (.*) ([^ ]+)',
                    '\\3, \\2 (\\1)'
                ],
                [
                    'Artist',
                    '(Orquesta Típica|Dúo|Cuarteto|Sexteto) (.*)',
                    '\\2, \\1'
                ],
                [
                    'Artist',
                    '(.*) ((?:Di|De) *[^ ]+)',
                    '\\2, \\1'
                ],
                [
                    'Artist',
                    '(.*) ([^ ]+)',
                    '\\2, \\1'
                ],
                [
                    'Artist',
                    '(.*)',
                    '\\1'
                ]
            ]))
        performer_last_name = data.FieldWithTable(
            'PerformerLastName', visible=False,
            table=[
                [
                    'Performer',
                    '^(.*) ((?:Di|De|Del) *[^ ]+)$',
                    '\\2'
                ],
                [
                    'Performer',
                    '^(.*) ([^ ]+)$',
                    '\\2'
                ],
                [
                    'Performer',
                    '^(.*)$',
                    '\\1'
                ]
            ])
        self.shared.songlist_fields.register_field(data.Field(
            'PerformersLastNames', visible=False,
            get_value=lambda song: ', '.join(performer_last_name.get_value({'Performer': name}) for name in song.get('Performer').split(', ')) if song.get('Performer') else None))

    def deactivate(self):
        del self.shared.songlist_fields



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, GLib, Gtk, Gdk
import logging

from .data import format_time
from .utils import action
from .utils.logger import logger


class WindowLoggingHandeler(logging.Handler):
    MAX_INFOBARS = 5

    def __init__(self, box, timeout):
        super(WindowLoggingHandeler, self).__init__(logging.INFO)
        self.box = box
        self.timeout = timeout

    def emit(self, record):
        children = self.box.get_children()
        nchildren = len(children)
        if nchildren >= self.MAX_INFOBARS:
            for child in children[:nchildren - self.MAX_INFOBARS + 1]:
                self.remove_infobar(child)
        message_type = Gtk.MessageType.ERROR if record.levelno >= 40 else Gtk.MessageType.WARNING if record.levelno >= 30 else Gtk.MessageType.INFO
        infobar = Gtk.InfoBar(visible=True, message_type=message_type)
        infobar.get_content_area().add(Gtk.Label(visible=True, label=record.msg))
        infobar.connect('response', self.remove_infobar)
        css_provider = Gtk.CssProvider.new()
        css_provider.load_from_data(b'infobar.error { background-image: linear-gradient(to top left, rgb(211,10,10), rgb(255,0,0)); }')
        infobar.get_style_context().add_provider(css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        self.box.add(infobar)
        infobar.timeout = GLib.timeout_add(self.timeout, self.remove_infobar, infobar)

    def remove_infobar(self, infobar, *args):
        GLib.source_remove(infobar.timeout)
        self.box.remove(infobar)

    def close(self):
        super(WindowLoggingHandeler, self).close()
        self.box.foreach(self.remove_infobar)


class Window(Gtk.ApplicationWindow):
    DEFAULT_TITLE = 'GAMPC'

    def __init__(self, app):
        super(Window, self).__init__(application=app)
        self.shared = app.shared
        self.is_fullscreen = False
        self.module = None

        self.config = self.shared.config.window
        self.config.access('message-timeout', 5)
        self.set_default_size(self.config.access('width', 1000),
                              self.config.access('height', 600))

        self.connect('destroy', self.destroy_cb)

        self.shared.ampd_server_properties.connect('notify::current-song', self.notify_current_song_cb)
        self.shared.ampd_server_properties.connect('notify::state', self.update_title)
        self.shared.connect('notify::server-label', self.update_subtitle)

        builder = self.shared.build_ui('window')

        self.option_buttons = {}
        for option in self.shared.ampd_server_properties.OPTION_NAMES:
            button = builder.get_object('togglebutton-' + option)
            self.shared.ampd_server_properties.bind_property('option-' + option, button, 'active', GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE)
            self.shared.bind_property('restricted', button, 'sensitive', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: not value)
            self.option_buttons[option] = button
            self.option_notify_active_cb(button, None)
            button.connect('notify::active', self.option_notify_active_cb)

        self.scale_time = builder.get_object('scale-time')
        label_time = builder.get_object('label-time')
        label_time_total = builder.get_object('label-time-total')
        self.volume_button = builder.get_object('volume-button')

        self.main = builder.get_object('main')
        self.add(self.main)
        self.titlebar = builder.get_object('header')
        self.set_titlebar(self.titlebar)

        self.shared.ampd_server_properties.bind_property('state', builder.get_object('button-play'), 'visible', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: value != 'play')
        self.shared.ampd_server_properties.bind_property('state', builder.get_object('button-pause'), 'visible', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: value == 'play')
        self.shared.ampd_server_properties.bind_property('state', self.scale_time, 'sensitive', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: value in {'play', 'pause'})
        self.shared.ampd_server_properties.bind_property('volume', self.volume_button, 'value', GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE)
        self.shared.ampd_server_properties.bind_property('volume', self.volume_button, 'sensitive', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: value != -1)
        self.shared.bind_property('restricted', builder.get_object('image-restricted'), 'visible', GObject.BindingFlags.SYNC_CREATE)
        self.shared.bind_property('restricted', self.scale_time, 'sensitive', GObject.BindingFlags.INVERT_BOOLEAN | GObject.BindingFlags.SYNC_CREATE)
        label_time.bind_property('visible', label_time_total, 'visible', GObject.BindingFlags.SYNC_CREATE)
        self.shared.ampd_server_properties.bind_property('bitrate', builder.get_object('label-bitrate'), 'label', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: ' [{kbps}kbps]'.format(kbps=value) if value else '')
        self.shared.ampd_server_properties.bind_property('time', self.scale_time.get_adjustment(), 'upper', GObject.BindingFlags.SYNC_CREATE)
        self.shared.ampd_server_properties.bind_property('time', self.scale_time.get_adjustment(), 'value', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: binding.get_source().elapsed)
        self.shared.ampd_server_properties.bind_property('elapsed', self.scale_time, 'fill-level', GObject.BindingFlags.SYNC_CREATE)
        self.scale_time_binding = None
        self.scale_time.get_adjustment().bind_property('value', label_time, 'label', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: format_time(value + 0.5))
        self.scale_time.get_adjustment().bind_property('upper', label_time_total, 'label', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: ' / ' + format_time(value))
        self.scale_time.get_adjustment().bind_property('upper', label_time, 'visible', GObject.BindingFlags.SYNC_CREATE, lambda binding, value: value != 0.0)
        self.set_scale_time_binding()

        self.add_action(action.Action('toggle-fullscreen', self.action_toggle_fullscreen_cb))
        self.add_action(action.Action('volume-popup', self.action_volume_popup_cb))

        builder.connect_signals(self)

        self.logging_handler = WindowLoggingHandeler(builder.get_object('box-info'), self.config.message_timeout * 1000)
        logger.addHandler(self.logging_handler)
        self.update_title()
        self.update_subtitle()

    @staticmethod
    def destroy_cb(self):
        logger.debug("Destroying window: {}".format(self))
        self.change_module(None)
        logger.removeHandler(self.logging_handler)
        self.logging_handler.close()
        self.shared.disconnect_by_func(self.update_subtitle)
        self.shared.ampd_server_properties.disconnect_by_func(self.update_title)
        self.shared.ampd_server_properties.disconnect_by_func(self.notify_current_song_cb)

    def option_notify_active_cb(self, button, param):
        option_name = '{name} mode'.format(name=button.get_name().capitalize())
        button.set_tooltip_text('{} {}'.format(_(option_name), _("on") if button.get_active() else _("off")))

    def change_module(self, module):
        if self.module:
            self.main.remove(self.module)
            self.module.disconnect_by_func(self.update_subtitle)
        self.module = module
        if module:
            self.main.pack_start(module, True, True, 0)
            self.module.connect('notify::full-title', self.update_subtitle)
        self.update_subtitle()

    def notify_current_song_cb(self, *args):
        self.scale_time_break_interaction()
        self.update_title()

    def update_title(self, *args):
        if self.shared.ampd_server_properties.state == 'stop':
            title = self.DEFAULT_TITLE
        else:
            title = '{} / {}'.format(self.shared.ampd_server_properties.current_song.get('Artist', ''), self.shared.ampd_server_properties.current_song.get('Title', ''))
            if self.shared.ampd_server_properties.state == 'pause':
                title += ' ({})'.format(_("paused"))
        self.titlebar.set_title(title)

    def update_subtitle(self, *args):
        components = []
        if self.module:
            components.append(self.module.full_title)
        if self.shared.server_label:
            components.append(self.shared.server_label)
        self.titlebar.set_subtitle(' / '.join(components))

    def do_check_resize(self):
        if not self.is_fullscreen:
            self.config.width, self.config.height = self.get_size()
        Gtk.ApplicationWindow.do_check_resize(self)

    def scale_time_button_press_event_cb(self, scale_time, event):
        if event.button == 1 and event.type == Gdk.EventType.BUTTON_PRESS and self.scale_time_binding:
            self.scale_time_binding.unbind()
            self.scale_time_binding = None
        else:
            self.scale_time_break_interaction()

    def scale_time_button_release_event_cb(self, scale_time, event):
        if not self.scale_time_binding:
            if event.button == 1 and event.state & (Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.MOD1_MASK) == 0:
                self.shared.ampd_server_properties.elapsed = self.scale_time.get_value()
                self.set_scale_time_binding()
            else:
                self.scale_time_break_interaction()

    def scale_time_break_interaction(self, *args):
        if not self.scale_time_binding:
            self.scale_time.set_sensitive(False)
            GLib.timeout_add(0, self.scale_time.set_sensitive, True)

    def scale_time_notify_sensitive_cb(self, *args):
        if not self.scale_time.get_sensitive():
            self.set_scale_time_binding()

    def set_scale_time_binding(self):
        if not self.scale_time_binding:
            self.scale_time_binding = self.shared.ampd_server_properties.bind_property('elapsed', self.scale_time.get_adjustment(), 'value', GObject.BindingFlags.SYNC_CREATE)

    def action_toggle_fullscreen_cb(self, *args):
        if self.is_fullscreen:
            self.unfullscreen()
        else:
            self.fullscreen()

    def action_volume_popup_cb(self, action, parameter):
        if self.volume_button.is_sensitive() and not self.volume_button.get_popup().get_mapped():
            self.volume_button.emit('popup')
        else:
            self.volume_button.emit('popdown')

    def do_window_state_event(self, event):
        self.is_fullscreen = bool(event.new_window_state & Gdk.WindowState.FULLSCREEN)
        self.set_show_menubar(not self.is_fullscreen)
        self.titlebar.set_visible(not self.is_fullscreen)
        return Gtk.ApplicationWindow.do_window_state_event(self, event)



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk

import ampd

from gampc import module
from gampc import plugin


class Command(module.Module):
    title = _("Execute MPD commands")
    name = 'command'
    key = '7'

    def __init__(self, *args):
        super(Command, self).__init__(*args)
        self.label = Gtk.Label(max_width_chars=50, wrap=True, selectable=True, visible=True)
        self.entry = Gtk.Entry(visible=True)
        self.entry.connect('activate', self.entry_activate_cb)
        self.entry.connect('focus-in-event', self.entry_focus_cb)
        self.entry.connect('focus-out-event', self.entry_focus_cb)
        widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True)
        widget.pack_start(self.label, True, True, 0)
        widget.pack_end(self.entry, False, False, 0)
        self.add(widget)
        self.connect('map', lambda self: self.entry.grab_focus())

    @ampd.task
    async def entry_activate_cb(self, entry):
        reply = await self.ampd._raw(entry.get_text())
        self.label.set_label('\n'.join(str(x) for x in reply) if reply else _("Empty reply"))

    def entry_focus_cb(self, entry, event):
        self.shared.enable_fragile_accels = not event.in_
        return False


class CommandExtension(plugin.Extension):
    modules = [Command]



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import os


root = logging.getLogger()
root.setLevel(os.environ.get('LOGLEVEL', 'INFO'))

handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter(fmt='%(levelname)s: %(name)s: %(message)s (%(pathname)s %(lineno)d)'))
root.addHandler(handler)

logger = logging.getLogger(__name__.split('.')[0])



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gio, GLib


class Item(object):
    def __init__(self, path, target, label=None, accels=None, **attributes):
        """
        path = NAME/NAME/..../NAME
        target = Gio.MENU_LINK_SUBMENU | Gio.MENU_LINK_SECTION | detailed-action
        """

        self.path = path
        self.target = target
        self.label = label
        self.attributes = attributes
        self.attributes.setdefault('hidden-when', 'action-missing')
        if accels:
            self.attributes['accels'] = accels
            self.attributes['detailed-action'] = target

    def __repr__(self):
        return "{}({})".format(self.path, self.target)


class OrderedMenu(Gio.Menu):
    ATTRIBUTE_NAME = 'omenu-name'

    def __init__(self, _set_accels=None):
        super(OrderedMenu, self).__init__()
        self._set_accels = _set_accels

    def insert_path(self, path, target, label=None, accels=None, **attributes):
        return self.insert(Item(path, target, label, accels, **attributes))

    def insert(self, item):
        menu, i, found, name = self._find_path(item.path.split('/'))
        if found:
            if item.target in (Gio.MENU_LINK_SUBMENU, Gio.MENU_LINK_SECTION):
                submenu = menu.get_item_link(i, item.target)
                if submenu:
                    return submenu
            raise ValueError("Item already exists for path '{path}'".format(path=item.path))
        if self._set_accels and 'detailed-action' in item.attributes:
            self._set_accels(item.attributes['detailed-action'], item.attributes.get('accels', []))
        return menu._create_item(i, name, item.target, item.label, item.attributes)

    def remove_path(self, path):
        menu, i, found, name = self._find_path(path.split('/'))
        if found:
            action = menu.get_item_attribute_value(i, 'detailed-action')
            if action and self._set_accels:
                self._set_accels(action.unpack(), [])
            menu.remove(i)
        else:
            raise ValueError("Cannot remove inexistent path '{path}'".format(path=path))

    def enable_fragile_accels(self):
        for action, accels, fragile in self._action_accels():
            if fragile:
                self._set_accels(action, accels)

    def disable_fragile_accels(self):
        for action, accels, fragile in self._action_accels():
            if fragile:
                self._set_accels(action, [])

    def _create_item(self, i, name, target, label=None, attributes=None):
        item = Gio.MenuItem.new(label)
        item.set_attribute_value(self.ATTRIBUTE_NAME, GLib.Variant.new_string(name))
        if target in (Gio.MENU_LINK_SUBMENU, Gio.MENU_LINK_SECTION):
            submenu = OrderedMenu()
            item.set_link(target, submenu)
        else:
            submenu = None
            item.set_detailed_action(target)
            if attributes:
                for attribute, value in attributes.items():
                    if value is None:
                        continue
                    elif isinstance(value, str):
                        value = GLib.Variant.new_string(value)
                    elif isinstance(value, bool):
                        value = GLib.Variant.new_boolean(value)
                    else:
                        value = GLib.Variant.new_strv(value)
                    item.set_attribute_value(attribute.replace('_', '-'), value)
        self.insert_item(i, item)
        return submenu

    def _find_position(self, name):
        n = self.get_n_items()
        for i in range(n):
            item_name = self.get_item_attribute_value(i, self.ATTRIBUTE_NAME)
            item_name = item_name and item_name.get_string()
            if item_name is None or item_name >= name:
                return i, item_name == name
        return n, False

    def _find_path(self, names):
        menu = self
        last_name = names[0]
        i, found = self._find_position(last_name)
        for name in names[1:]:
            if not found:
                menu = menu._create_item(i, last_name, Gio.MENU_LINK_SECTION)
            else:
                menu = menu.get_item_link(i, Gio.MENU_LINK_SUBMENU) or menu.get_item_link(i, Gio.MENU_LINK_SECTION)
                if not menu:
                    raise ValueError("Path element '{name}' is not a submenu or section".format(name=last_name))
            i, found = menu._find_position(name)
            last_name = name

        return menu, i, found, last_name

    def _action_accels(self):
        for i in range(self.get_n_items()):
            submenu = self.get_item_link(i, Gio.MENU_LINK_SUBMENU) or self.get_item_link(i, Gio.MENU_LINK_SECTION)
            if submenu:
                yield from submenu._action_accels()
            else:
                item_accels = self.get_item_attribute_value(i, 'accels')
                item_accels_fragile = self.get_item_attribute_value(i, 'accels-fragile')
                if item_accels:
                    yield self.get_item_attribute_value(i, 'detailed-action').unpack(), item_accels.unpack(), item_accels_fragile and item_accels_fragile.unpack()



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GLib
from gi.repository import Gtk

import ampd

from gampc import plugin
from gampc.utils import action
from gampc.utils import omenu
from gampc.utils import ssde

import songlist


class PlayQueue(songlist.SongListWithTotals):
    title = _("Play Queue")
    name = 'playqueue'
    key = '1'

    duplicate_test_columns = ['Title']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.actions.add_action(action.Action('playqueue-replace-playlist', self.action_replace_playlist_cb))
        self.actions.add_action(action.Action('playqueue-shuffle', self.action_shuffle_cb, restrict=True, shared=self.shared))
        self.actions.add_action(action.Action('playqueue-add-url', self.action_add_url_cb))
        self.actions.add_action(action.Action('playqueue-go-to-current', self.action_go_to_current_cb))
        self.signal_handler_connect(self.shared.ampd_server_properties, 'notify::current-song', self.notify_current_song_cb)
        for name in self.actions.list_actions():
            if name.startswith('playqueue-ext-'):
                self.actions.remove(name)

    @ampd.task
    async def client_connected_cb(self, client):
        while True:
            self.set_songs(await self.ampd.playlistinfo())
            await self.ampd.idle(ampd.PLAYLIST)

    def data_func(self, column, renderer, store, i, j):
        super().data_func(column, renderer, store, i, j)
        if self.shared.ampd_server_properties.state != 'stop' and store.get_record(i).Id == self.shared.ampd_server_properties.current_song.get('Id'):
            renderer.set_property('font', 'italic bold')
            bg = self._mix_colors(0, 0, 0)

            renderer.set_property('background-rgba', bg)
        else:
            renderer.set_property('font', None)

    @ampd.task
    async def action_replace_playlist_cb(self, action, parameter):
        playlists = sorted(map(lambda x: x['playlist'], await self.ampd.listplaylists()))
        if not playlists:
            return
        struct = ssde.Choice(playlists, label=_("Replace a playlist with play queue"))
        value = struct.dialog(self.win)
        if not value:
            return
        await self.ampd.command_list((self.ampd.rm(value), self.ampd.save(value)))

    @ampd.task
    async def action_shuffle_cb(self, action, parameter):
        await self.ampd.shuffle()

    def action_go_to_current_cb(self, action, parameter):
        if self.shared.ampd_server_properties.current_song:
            p = Gtk.TreePath.new_from_string(self.shared.ampd_server_properties.current_song['Pos'])
            self.treeview.set_cursor(p)
            self.treeview.scroll_to_cell(p, None, True, 0.5, 0.0)

    def notify_current_song_cb(self, *args):
        self.treeview.queue_draw()

    def new_song(self, store, i):
        ampd.Task(self.ampd.addid(store.get_record(i).file, store.get_path(i).get_indices()[0]))

    def delete_song(self, store, i):
        song_id = store.get_record(i).Id
        m = int(store.get_string_from_iter(i))
        for n in range(store.iter_n_children()):
            if n != m and store.get_record(store.iter_nth_child(None, n)).Id == song_id:
                ampd.Task(self.ampd.command_list((self.ampd.swap(n, m), self.ampd.delete(m))))
                store.remove(i)
                return
        if not (self.shared.restricted and self.shared.ampd_server_properties.state == 'play' and self.shared.ampd_server_properties.current_song.get('pos') == song_id):
            ampd.Task(self.ampd.deleteid(song_id))
            store.remove(i)

    @ampd.task
    async def treeview_row_activated_cb(self, treeview, p, column):
        if not self.shared.restricted:
            await self.ampd.playid(self.store.get_record(self.store.get_iter(p)).Id)

    @ampd.task
    async def action_add_url_cb(self, action, parameter):
        await self.ampd.ping()
        struct = ssde.Text(label=_("URL to add"), default='http://')
        url = struct.dialog(self.win)
        if url:
            await self.ampd.add(url)


@ampd.task
async def action_playqueue_add_replace_cb(songlist, action, parameter):
    filenames = songlist.get_filenames(not parameter.get_boolean())
    replace = '-replace' in action.get_name()
    shuffle = '-shuffle' in action.get_name()
    if replace:
        await songlist.ampd.clear()
    await songlist.ampd.command_list(songlist.ampd.add(filename) for filename in filenames)
    if shuffle:
        await songlist.ampd.shuffle()
    if replace:
        await songlist.ampd.play()


class PlayQueueExtension(plugin.Extension):
    modules = [PlayQueue]

    def activate(self):
        self.provides['app'] = {}
        self.provides['app']['menubar_items'] = [
            omenu.Item('20#edit/90#playqueue/10', 'mod.playqueue-replace-playlist', _("Replace a playlist with play queue")),
            omenu.Item('20#edit/90#playqueue/20', 'mod.playqueue-shuffle', _("Shuffle")),
            omenu.Item('20#edit/90#playqueue/30', 'mod.playqueue-add-url', _("Add URL")),
            omenu.Item('20#edit/90#playqueue/40', 'mod.playqueue-go-to-current', _("Go to current song"), ['<Control>z'])
        ]

        self.provides['songlist'] = {}
        self.provides['songlist']['actions'] = [
            action.ActionModel('playqueue-ext' + verb + shuffle, action_playqueue_add_replace_cb,
                               restrict=(verb == '-replace' or shuffle), parameter_type=GLib.VariantType.new('b'), shared=self.shared)
            for verb in ('-add', '-replace') for shuffle in ('', '-shuffle')
        ]
        for name, parameter in (('context_menu_items', '(false)'), ('left_context_menu_items', '(true)')):
            self.provides['songlist'][name] = [
                omenu.Item('10/1', 'mod.playqueue-ext-add' + parameter, _("Add to play queue")),
                omenu.Item('10/2', 'mod.playqueue-ext-replace' + parameter, _("Replace play queue")),
                omenu.Item('15/1', 'mod.playqueue-ext-add-shuffle' + parameter, _("Add to play queue & shuffle")),
                omenu.Item('15/2', 'mod.playqueue-ext-replace-shuffle' + parameter, _("Replace play queue & shuffle")),
            ]

        self.provides[PlayQueue.name] = {}
        self.provides[PlayQueue.name]['context_menu_items'] = [
            omenu.Item('90/10', 'mod.playqueue-shuffle', _("Shuffle")),
            omenu.Item('90/20', 'mod.playqueue-add-url', _("Add URL")),
        ]



# coding: utf-8
#
# Auto-generated Gtk3 Simple Structured Data Editor
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, Gdk, Gtk


class InvalidData(Exception):
    pass


class EditorRow(GObject.Object):
    def __init__(self, struct, draw_label, has_value=False):
        super(EditorRow, self).__init__()
        self.struct = struct
        self.draw_label = draw_label
        self.has_value = has_value
        self.subrows = []


class DialogEditor(Gtk.Dialog):
    valid = GObject.property(type=bool, default=True)
    editing = GObject.property(type=bool, default=False)

    def __init__(self, struct, parent, value=None, size=None, scrolled=False):
        super(DialogEditor, self).__init__(parent=parent)
        self._struct = struct
        self.set_destroy_with_parent(True)
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.ok_button = self.add_button(_("_OK"), Gtk.ResponseType.OK)
        self.bind_property('valid', self.ok_button, 'sensitive', GObject.BindingFlags.SYNC_CREATE)
        self.bind_property('editing', self.ok_button, 'sensitive', GObject.BindingFlags.INVERT_BOOLEAN)

        self._value_col = Gtk.TreeViewColumn.new()

        self._renderers = {}
        for target in set(map(type, self._struct._get_all_targets())):
            renderer = target.get_renderer()
            renderer.handlers = [renderer.connect('editing-started', self.editing_started_cb)]
            if target.edited_signal:
                renderer.handlers.append(renderer.connect(target.edited_signal, target.edited_cb, self))
            self._value_col.pack_start(renderer, False)
            self._value_col.set_cell_data_func(renderer, self._data_func, target)
            self._renderers[target] = renderer

        self._store = Gtk.TreeStore(EditorRow)
        self._treeview = Gtk.TreeView(visible=True, headers_visible=False, show_expanders=True, enable_tree_lines=True, enable_search=False, model=self._store)
        self._treeview.connect('key-press-event', self._key_press_event_cb)
        self._treeview.connect('row-activated', self.row_activated_cb)
        self._treeview.insert_column_with_data_func(0, '', Gtk.CellRendererText(), self._label_data_func)
        self._treeview.append_column(self._value_col)

        if size:
            self.set_default_size(*size)
        if scrolled:
            scrolled_window = Gtk.ScrolledWindow(visible=True, expand=True)
            scrolled_window.add(self._treeview)
            self.get_content_area().add(scrolled_window)
        else:
            self.get_content_area().add(self._treeview)
        if self._struct.label:
            self.set_title(self._struct.label)

        self._row = self._struct._form_row(value)
        self._set_row(None, self._row)
        self.value = None
        self.validate()

        reply = self.run()
        if reply != Gtk.ResponseType.OK:
            self.value = None

        if size:
            (size[0], size[1]) = self.get_size()
        self.destroy()
        for renderer in self._renderers.values():
            for handler in renderer.handlers:
                renderer.disconnect(handler)
        del self._value_col
        if parent:
            parent.present()

    def editing_started_cb(self, renderer, editable, path):
        editable.connect('editing-done', self.editing_done_cb, self._get_row(self._store.get_iter(path)))
        self.editing = True

    def editing_done_cb(self, editable, row):
        self.editing = False
        if not editable.get_property('editing-canceled'):
            row.struct._editing_done(editable, self, row)
            self.validate()

    def validate(self):
        try:
            self.value = self._struct._get_value(self._row)
            self.valid = True
        except:
            self.valid = False
            raise

    def _data_func(self, column, renderer, store, i, target):
        row = self._get_row(i)
        if isinstance(row.struct, target):
            renderer.set_visible(True)
            row.struct._data_func(renderer, row)
        else:
            renderer.set_visible(False)

    def _label_data_func(self, column, renderer, store, i):
        row = self._get_row(i)
        renderer.set_property('text', row.struct.label if row.draw_label else None)

    def _key_press_event_cb(self, treeview, event):
        path, column = self._treeview.get_cursor()
        if not path:
            return False
        j = self._store.get_iter(path)

        if event.state & Gdk.ModifierType.MOD1_MASK:
            k = self._store.iter_previous(j) if event.keyval == Gdk.KEY_Up else self._store.iter_next(j) if event.keyval == Gdk.KEY_Down else None
            if k:
                self._treeview.set_cursor(self._store.get_path(k))
            return True

        i = self._store.iter_parent(j)
        parent = self._get_row(i) if i else self._row
        row = self._get_row(j)
        index = -1 if not i and row == parent else parent.subrows.index(row)
        reply = parent.struct._key_press(parent, index, event)
        if reply:
            self._set_subrows(parent)
            if reply == event:
                self._key_press_event_cb(treeview, event)
            return True
        else:
            return False

    def row_activated_cb(self, treeview, path, column):
        renderer = self._renderers.get(type(self._get_row(self._store.get_iter(path)).struct))
        if renderer:
            treeview.set_cursor_on_cell(path, self._value_col, renderer, True)
        elif treeview.row_expanded(path):
            treeview.collapse_row(path)
        else:
            treeview.expand_row(path, True)

    def _get_row(self, i):
        return self._store.get_value(i, 0)

    def _set_row(self, i, row):
        if not i and row.has_value:
            i = self._store.append(None)
            row.draw_label = False
        if i:
            self._store.set_value(i, 0, row)
            row.path = self._store.get_path(i)
        else:
            row.path = None
        self._set_subrows(row)

    def _set_subrows(self, row):
        i = row.path and self._store.get_iter(row.path)
        j = self._store.iter_children(i)
        for subrow in row.subrows:
            if j:
                self._set_row(j, subrow)
                j = self._store.iter_next(j)
            else:
                self._set_row(self._store.append(i), subrow)
        if j:
            while self._store.remove(j):
                pass
        if row.path:
            self._treeview.expand_row(row.path, True)
        self.validate()


class _Struct(object):
    def __init__(self, default=None, label=None, name=None, validator=lambda value: True):
        self.default = default
        self.label = label
        self.name = name or label
        self.validator = validator
        self._substructs = []

    def dialog(*args, **kwargs):
        return DialogEditor(*args, **kwargs).value

    def _get_all_targets(self):
        return [self] if isinstance(self, _Target) else sum((substruct._get_all_targets() for substruct in self._substructs), [])

    def _get_value(self, row):
        raise NotImplementedError

    @staticmethod
    def _key_press(row, index, event):
        return False


class _Target(_Struct):
    edited_signal = None

    def _form_row(self, value=None, draw_label=True, has_value=True):
        row = EditorRow(self, draw_label, has_value)
        row.value = value or self.default
        return row

    def _get_value(self, row):
        return row.value


class Text(_Target):
    @staticmethod
    def get_renderer():
        return Gtk.CellRendererText(editable=True)

    @staticmethod
    def _data_func(renderer, row):
        renderer.set_property('text', row.value)

    @staticmethod
    def _editing_done(editable, editor, row):
        row.value = editable.get_text()


class Integer(_Target):
    def __init__(self, default=None, *args, **kwargs):
        super(Integer, self).__init__(default or 0, *args, **kwargs)

    @staticmethod
    def get_renderer():
        return Gtk.CellRendererSpin(editable=True, adjustment=Gtk.Adjustment(lower=-GObject.G_MAXDOUBLE, upper=GObject.G_MAXDOUBLE, step_increment=1))

    @staticmethod
    def _data_func(renderer, row):
        renderer.set_property('text', str(row.value))

    @staticmethod
    def _editing_done(editable, editor, row):
        try:
            row.value = int(editable.get_text())
        except:
            pass


class Boolean(_Target):
    edited_signal = 'toggled'

    def __init__(self, default=None, *args, **kwargs):
        super(Boolean, self).__init__(default or False, *args, **kwargs)

    @staticmethod
    def get_renderer():
        return Gtk.CellRendererToggle(xalign=0.0)

    @staticmethod
    def _data_func(renderer, row):
        renderer.set_property('active', row.value)

    @staticmethod
    def edited_cb(renderer, ps, editor):
        row = editor._get_row(editor._store.get_iter_from_string(ps))
        row.value = not row.value
        editor.validate()


class Choice(_Target):
    def __init__(self, choices, default=None, *args, **kwargs):
        if not choices:
            raise ValueError
        if not isinstance(default, tuple):
            default = (default, None)
        elif len(default) != 2:
            raise ValueError
        super(Choice, self).__init__(default, *args, **kwargs)
        self.default_choice = default[0]
        self.choices = {}
        self.choice_store = Gtk.ListStore()
        self.choice_store.set_column_types([str])
        for choice in choices:
            if isinstance(choice, str):
                name = choice
                self.choices[name] = None
            elif isinstance(choice, _Struct) and choice.label:
                name = choice.label
                self.choices[name] = choice
                self._substructs.append(choice)
            else:
                raise ValueError
            self.choice_store.set_value(self.choice_store.append(), 0, name)
            self.default_choice = self.default_choice or name
        if self.default_choice not in self.choices:
            raise ValueError

    def _form_row(self, value=None, draw_label=True, has_value=True):
        row = EditorRow(self, draw_label, has_value)
        value = value or self.default
        row.name, subvalue = value if isinstance(value, tuple) else (value, None)
        if row.name not in self.choices:
            row.name, subvalue = self.default_choice, None
        row.choice = self.choices[row.name]
        if row.choice:
            row.subrows = [row.choice._form_row(subvalue)]
        return row

    def _get_value(self, row):
        return (row.name, row.choice._get_value(row.subrows[0])) if row.choice else row.name

    @staticmethod
    def get_renderer():
        return Gtk.CellRendererCombo(editable=True, text_column=0, has_entry=False)

    def _data_func(self, renderer, row):
        renderer.set_property('model', self.choice_store)
        renderer.set_property('text', row.name)

    def _editing_done(self, combo_box, editor, row):
        j = combo_box.get_active_iter()
        row.name = self.choice_store.get_value(j, 0)
        row.choice = self.choices[row.name]
        if row.choice:
            row.subrows = [row.choice._form_row()]
        else:
            row.subrows = []
        editor._set_subrows(row)


class Tuple(_Struct):
    def __init__(self, substructs, *args, **kwargs):
        super(Tuple, self).__init__(*args, **kwargs)
        self._substructs = substructs

    def _get_nth_substruct(self, n):
        return self._substructs[n]

    def _form_row(self, value=None, draw_label=True, has_value=False):
        row = EditorRow(self, draw_label, has_value)
        row.subrows = [substruct._form_row(subvalue) for substruct, subvalue in self._list_substruct_value(value)]
        return row

    def _list_substruct_value(self, value):
        return ((self._get_nth_substruct(j), (value[j] if value else None)) for j in range(len(self._substructs)))

    def _get_value(self, row):
        return [subrow.struct._get_value(subrow) for subrow in row.subrows]


class Dict(Tuple):
    def _list_substruct_value(self, value):
        return ((substruct, (value.get(substruct.name) if value else None)) for substruct in self._substructs)

    def _get_value(self, row):
        return {subrow.struct.name: subrow.struct._get_value(subrow) for subrow in row.subrows}


def _list_key_press(row, index, event, substruct, base=0):
    if index >= base and event.state & Gdk.ModifierType.CONTROL_MASK:
        if event.keyval == Gdk.KEY_plus:
            row.subrows.insert(index, substruct._form_row())
        elif event.keyval == Gdk.KEY_minus and len(row.subrows) > base + 1:
            row.subrows.pop(index)
        elif event.keyval == Gdk.KEY_Up or event.keyval == Gdk.KEY_Down:
            if event.keyval == Gdk.KEY_Up and index > base:
                index2 = index - 1
            elif event.keyval == Gdk.KEY_Down and index < len(row.subrows) - 1:
                index2 = index + 1
            else:
                return True
            row.subrows[index], row.subrows[index2] = row.subrows[index2], row.subrows[index]
            event.state &= ~Gdk.ModifierType.CONTROL_MASK
            event.state |= Gdk.ModifierType.MOD1_MASK
            return event
        return True
    return False


class List(Tuple):
    def __init__(self, substruct, *args, **kwargs):
        super(List, self).__init__([substruct], *args, **kwargs)
        self._substruct = substruct

    def _get_nth_substruct(self, n):
        return self._substruct

    def _list_substruct_value(self, value):
        return ((self._substruct, subvalue) for subvalue in value or [None])

    def _key_press(self, row, index, event):
        return _list_key_press(row, index, event, self._substruct)


class TupleAndList(Tuple):
    def __init__(self, substructs, *args, **kwargs):
        super(TupleAndList, self).__init__(substructs, *args, **kwargs)
        self.N = len(substructs) - 1
        self._list_substruct = substructs[-1]

    def _get_nth_substruct(self, n):
        return self._substructs[n] if n < self.N else self._list_substruct

    def _list_substruct_value(self, value):
        return Tuple._list_substruct_value(self, value[:self.N] + value[self.N] if value else None)

    def _get_value(self, row):
        value = Tuple._get_value(self, row)
        return value[:self.N] + [value[self.N:]]

    def _key_press(self, row, index, event):
        return _list_key_press(row, index, event, self._list_substruct, self.N)



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk

import ampd

from gampc import module
from gampc import plugin
from gampc.utils import action
from gampc.utils import omenu
from gampc.utils import ssde

import songlist


class Playlist(songlist.EditableSongList, songlist.SongListWithTotals, module.PanedModule):
    title = _("Playlist Editor")
    name = 'playlist'
    key = '5'

    duplicate_test_columns = ['file']

    def __init__(self, *args, **kwargs):
        super(Playlist, self).__init__(*args, **kwargs)

        self.actions.add_action(action.Action('playlist-rename', self.action_playlist_rename_cb))
        self.actions.add_action(action.Action('playlist-delete', self.action_playlist_delete_cb))

        self.playlist_name = None
        self.modified = False

        self.left_store = Gtk.ListStore()
        self.left_store.set_column_types([str])
        self.left_treeview.set_model(self.left_store)
        self.left_treeview.insert_column_with_attributes(0, '', Gtk.CellRendererPixbuf(icon_name='view-list-symbolic'))
        self.left_treeview.insert_column_with_data_func(1, _("Playlist name"), Gtk.CellRendererText(), self.playlist_name_data_func)
        self.left_treeview.connect('row-activated', self.left_treeview_row_activated_cb)

    def playlist_name_data_func(self, column, renderer, store, i):
        name = store.get_value(i, 0)
        if name == self.playlist_name and self.modified:
            renderer.set_property('text', '* ' + name)
            renderer.set_property('font', 'bold italic')
        else:
            renderer.set_property('text', name)
            renderer.set_property('font', None)

    @ampd.task
    async def client_connected_cb(self, client):
        while True:
            playlists = sorted(map(lambda x: x['playlist'], await self.ampd.listplaylists()))
            self.left_store_set_rows(playlists)
            p = Gtk.TreePath.new_from_indices([playlists.index(self.playlist_name)]) if self.playlist_name in playlists else Gtk.TreePath.new_first()
            if (self.left_treeview.get_selection().path_is_selected(p)):
                self.left_treeview_selection_changed_cb()
            else:
                self.left_treeview.set_cursor(p)
            await self.ampd.idle(ampd.STORED_PLAYLIST)

    def action_reset_cb(self, action, parameter):
        super().action_reset_cb(action, parameter)
        self.load_playlist()

    def left_treeview_selection_changed_cb(self, *args):
        store, i = self.left_treeview.get_selection().get_selected()
        self.playlist_name = store.get_value(i, 0) if i else None
        self.load_playlist()

    def left_treeview_row_activated_cb(self, left_treeview, p, col):
        self.action_playlist_rename_cb(None, None)

    @ampd.task
    async def load_playlist(self):
        self.set_modified(False)
        if not self.playlist_name:
            self.set_songs([])
            return
        songs = await self.ampd.listplaylistinfo(self.playlist_name)
        self.set_songs(songs)
        self.treeview.get_selection().unselect_all()

    def set_modified(self, modified=True):
        if modified != self.modified:
            self.modified = modified
            self.left_treeview.queue_draw()
        self.treeview.queue_draw()

    @ampd.task
    async def action_save_cb(self, action, parameter):
        tempname = '$$TEMP$$'
        try:
            await self.ampd.rm(tempname)
        except:
            pass
        try:
            await self.ampd.command_list([self.ampd.playlistadd(tempname, song.file) for i, p, song in self.store if song._status != self.SONG_DELETED] + [self.ampd.rm(self.playlist_name), self.ampd.rename(tempname, self.playlist_name)])
        except:
            try:
                await self.ampd.rm(tempname)
            except:
                pass
            raise
        self.treeview.get_selection().unselect_all()
        self.load_playlist()

    @ampd.task
    async def action_playlist_rename_cb(self, action, parameter):
        struct = ssde.Text(label=_("Rename playlist"), default=self.playlist_name)
        new_name = struct.dialog(self.win)
        if new_name and new_name != self.playlist_name:
            await self.ampd.rename(self.playlist_name, new_name)
            self.playlist_name = new_name

    @ampd.task
    async def action_playlist_delete_cb(self, action, parameter):
        if not self.playlist_name:
            return
        dialog = Gtk.Dialog(parent=self.win, title=_("Delete playlist"))
        dialog.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.get_content_area().add(Gtk.Label(label=_("Delete playlist {name}?").format(name=self.playlist_name), visible=True))
        reply = dialog.run()
        dialog.destroy()
        if reply != Gtk.ResponseType.OK:
            return
        await self.ampd.rm(self.playlist_name)


@ampd.task
async def action_playlist_add_cb(songlist_, action, parameter):
    filenames = songlist_.get_filenames()
    playlists = sorted(map(lambda x: x['playlist'], (await songlist_.ampd.listplaylists())))
    new_playlist = _("<New playlist>")
    struct = ssde.Choice([new_playlist] + playlists,
                         label=_("Add to playlist"))
    value = struct.dialog(songlist_.win)
    if value == new_playlist:
        new_name = _("<New playlist name>")
        struct = ssde.Text(label=_("New playlist name"),
                           default=new_name,
                           validator=lambda x: x != new_name)
        value = struct.dialog(songlist_.win)
    if not value:
        return
    await songlist_.ampd.command_list(songlist_.ampd.playlistadd(value, filename) for filename in filenames)


class PlaylistExtension(plugin.Extension):
    modules = [Playlist]

    def activate(self):
        self.provides['songlist'] = {}
        self.provides['songlist']['actions'] = [action.ActionModel('playlist-add', action_playlist_add_cb)]
        self.provides['songlist']['context_menu_items'] = [omenu.Item('20/10', 'mod.playlist-add', _("Add to playlist"))]

        self.provides[Playlist.name] = {}
        self.provides[Playlist.name]['left_context_menu_items'] = [
            omenu.Item('50/10', 'mod.playlist-rename', _("Rename playlist")),
            omenu.Item('50/20', 'mod.playlist-delete', _("Delete playlist"))
        ]



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Peas
import sys
import os
import xdg.BaseDirectory
import asyncio

import ampd

from .utils import db
from .utils import config
from .utils.logger import logger


class AppShared(GObject.GObject):
    want_to_connect = GObject.property(type=bool, default=False)
    restricted = GObject.property(type=bool, default=False)
    dark = GObject.property(type=bool, default=False)
    enable_fragile_accels = GObject.property(type=bool, default=True)
    server_label = GObject.property(type=str, default='')
    server_profile = GObject.property(type=str)

    CONFIG_EDIT_DIALOG_SIZE = 'edit-dialog-size'
    SEPARATOR_FILE = 'separator.mp3'
    STICKER_PROPERTIES = ('restricted', 'dark')

    def __init__(self):
        super(AppShared, self).__init__()
        self.uidefs = {}
        self.excepthook_orig, sys.excepthook = sys.excepthook, self.excepthook

        self.ampd_client = ampd.ClientGLib()
        self.ampd_client.connect('client-connected', self.client_connected_cb)
        self.ampd_client.connect('client-disconnected', self.client_disconnected_cb)
        self.ampd_server_properties = ampd.ServerPropertiesGLib(self.ampd_client)
        self.ampd_server_properties.connect('server-error', self.server_error_cb)
        for option in self.ampd_server_properties.OPTION_NAMES:
            self.ampd_server_properties.connect('notify::option-' + option, self.notify_option_cb)
        self.ampd_server_properties.connect('notify::updating-db', self.set_server_label)

        self.ampd_host = None
        self.ampd_port = 6600
        self.ampd = self.ampd_client.executor.sub_executor()

        self.peas_engine = Peas.Engine.get_default()
        self.peas_engine.enable_loader('python3')
        self.peas_extension_set = Peas.ExtensionSet.new(self.peas_engine, Peas.Activatable, [])
        self.peas_engine.add_search_path(os.path.join(os.path.dirname(__file__), 'plugins'))
        self.peas_engine.rescan_plugins()

        self.config = config.ConfigDict.load('config')

        self.config.server.access('profiles',
                                  [
                                      {
                                          'name': _("Local host"),
                                          'host': 'localhost',
                                          'port': 6600,
                                      },
                                  ])
        self.connect('notify::server-profile', self.ampd_connect)
        self.server_profile = self.config.server.access('profile', _("Local host"))

        self.separator_song = {'file': self.SEPARATOR_FILE}
        self.connect('notify', self.notify_sticker_cb)
        # self.restricted = self.config.access('restricted', False)
        self.connect('notify::restricted', self.notify_restricted_cb)
        self.connect('notify::dark', self.notify_dark_cb)

    def __del__(self):
        logger.debug('Deleting {}'.format(self))

    def close(self):
        logger.debug("Closing shared")
        self.config.save()
        asyncio.ensure_future(self.ampd_client.close())
        sys.excepthook = self.excepthook_orig
        self.disconnect_by_func(self.notify_restricted_cb)
        self.disconnect_by_func(self.notify_dark_cb)
        self.disconnect_by_func(self.notify_sticker_cb)
        self.disconnect_by_func(self.ampd_connect)
        self.ampd_server_properties.disconnect_by_func(self.set_server_label)
        del self.peas_extension_set
        del self.ampd_client
        del self.ampd_server_properties

    def build_ui(self, uiname):
        if uiname in self.uidefs:
            uidef = self.uidefs[uiname]
        else:
            uidef = open(os.path.join(os.path.dirname(__file__), uiname + '.ui')).read()
            self.uidefs[uiname] = uidef
        builder = Gtk.Builder(translation_domain='gampc')
        builder.add_from_string(uidef)
        return builder

    def client_connected_cb(self, client):
        logger.info(_("Connected to {profile} [protocol version {protocol}]").format(profile=self.config.server.profile, protocol=self.ampd.get_protocol_version()))
        self.idle_database()
        self.idle_sticker()

    @ampd.task
    async def client_disconnected_cb(self, client, reason, message):
        self.separator_song.clear()
        self.separator_song['file'] = self.SEPARATOR_FILE
        if reason == client.DISCONNECT_RECONNECT:
            return
        elif reason == client.DISCONNECT_PASSWORD:
            logger.error(_("Invalid password for {}").format(self.config.server.profile))
            return
        elif reason == client.DISCONNECT_FAILED_CONNECT:
            logger.error(_("Connection to {} failed").format(self.config.server.profile))
            if message:
                logger.error(message)
        else:
            logger.info(_("Disconnected from {}").format(self.config.server.profile))
        if self.want_to_connect:
            reply = await self.ampd.idle(ampd.CONNECT, timeout=1)
            if reply & ampd.TIMEOUT:
                asyncio.ensure_future(self.ampd_client.connect_to_server(self.ampd_host, self.ampd_port))

    def server_error_cb(self, client, error):
        logger.error(_("Server error: {error}").format(error=error))

    @ampd.task
    async def notify_option_cb(self, properties, param):
        if self.restricted:
            option = param.name.split('-')[1]
            await getattr(self.ampd, option)(0)

    @staticmethod
    def notify_restricted_cb(self, param):
        if self.restricted:
            for option in self.ampd_server_properties.OPTION_NAMES:
                self.ampd_server_properties.set_property('option-' + option, False)

    @staticmethod
    def notify_dark_cb(self, param):
        css = Gtk.CssProvider.get_named('Adwaita', 'dark' if self.dark else None)
        Gtk.StyleContext.add_provider_for_screen(Gdk.Screen.get_default(), css, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    @staticmethod
    @ampd.task
    async def notify_sticker_cb(self, param):
        if param.name in self.STICKER_PROPERTIES:
            try:
                await self.ampd.sticker_set('song', self.SEPARATOR_FILE, param.name, str(self.get_property(param.name)))
            except ampd.errors.ReplyError:
                pass

    @ampd.task
    async def idle_database(self):
        while True:
            song = await self.ampd.find('file', self.SEPARATOR_FILE)
            if song:
                self.separator_song.update(song[0])
            else:
                GLib.timeout_add(0, self.separator_missing)
            await self.ampd.idle(ampd.DATABASE)
            logger.info(_("Database changed"))

    def separator_missing(self):
        dialog = Gtk.MessageDialog(message_type=Gtk.MessageType.WARNING,
                                   buttons=Gtk.ButtonsType.CLOSE,
                                   text=_("Some features require a file named '{separator}' at the music root directory.  Such a file, consisting of a three second silence, is provided.").format(separator=self.SEPARATOR_FILE))
        dialog.run()
        dialog.destroy()

    @ampd.task
    async def idle_sticker(self):
        while True:
            self.handler_block_by_func(self.notify_sticker_cb)
            try:
                properties = await self.ampd.sticker_list('song', self.SEPARATOR_FILE)
            except ampd.errors.ReplyError:
                properties = []
            for property_ in properties:
                key, value = property_.split('=', 1)
                if key in self.STICKER_PROPERTIES:
                    self.set_property(key, value == 'True')
            self.handler_unblock_by_func(self.notify_sticker_cb)
            await self.ampd.idle(ampd.STICKER)

    def ampd_connect(self, *args):
        self.want_to_connect = True
        self.config.server.profile = self.server_profile
        self.set_server_label()
        for profile in self.config.server.profiles:
            if profile['name'] == self.server_profile:
                self.ampd_host = profile['host']
                self.ampd_port = profile['port']
                asyncio.ensure_future(self.ampd_client.connect_to_server(self.ampd_host, self.ampd_port))
                return

    def ampd_disconnect(self, *args):
        self.want_to_connect = False
        asyncio.ensure_future(self.ampd_client.disconnect_from_server())
        self.set_server_label()

    def set_server_label(self, *args):
        if self.want_to_connect:
            self.server_label = self.config.server.profile
            if self.ampd_server_properties.updating_db:
                self.server_label += " [{}]".format(_("database update"))
        else:
            self.server_label = ''

    def excepthook(self, *args):
        logger.error(args[1], exc_info=args)
        try:
            del sys.last_type, sys.last_value, sys.last_traceback
        except:
            pass

    @staticmethod
    def sqlite_connection(name):
        connection = db.Connection(os.path.join(xdg.BaseDirectory.save_data_path('gampc'), name + '.sqlite'))
        connection.cursor().execute('PRAGMA FOREIGH_KEYS=ON')
        return connection

    def collect_plugin_provides(self, targets):
        provides = {}
        self.peas_extension_set.foreach(self.collect_plugin_provides_worker, targets, provides)
        return provides

    @staticmethod
    def collect_plugin_provides_worker(extension_set, info, extension, targets, provides):
        for target, target_provides in extension.provides.items():
            if target in targets:
                for key, value in target_provides.items():
                    key_provides = provides.setdefault(key, [])
                    key_provides += value



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gio, GLib, Gtk, Gdk, PeasGtk
import signal
import logging
import dbus
import asyncio
import gbulb.gtk
import ampd

from . import window
from . import shared
from . import __program_name__, __version__, __program_description__, __copyright__, __license__
from .utils import omenu
from .utils import action
from .utils import ssde
from .utils.logger import logger


class App(Gtk.Application):
    def __init__(self):
        super(App, self).__init__(application_id='begnac.gampc', flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

        self.add_main_option('list-actions', 0, GLib.OptionFlags.NONE, GLib.OptionArg.NONE, _("List application actions"), None)
        self.add_main_option('version', 0, GLib.OptionFlags.NONE, GLib.OptionArg.NONE, _("Display version"), None)
        self.add_main_option('copyright', 0, GLib.OptionFlags.NONE, GLib.OptionArg.NONE, _("Display copyright"), None)
        self.add_main_option('non-unique', ord('u'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, _("Do not start a unique instance"), None)
        self.add_main_option('debug', 0, GLib.OptionFlags.NONE, GLib.OptionArg.NONE, _("Debug messages"), None)
        self.add_main_option(GLib.OPTION_REMAINING, 0, GLib.OptionFlags.NONE, GLib.OptionArg.STRING_ARRAY, '', _("[ACTION...]"))

    def __del__(self):
        logger.debug('Deleting {}'.format(self))

    def do_handle_local_options(self, options):
        if options.contains('non-unique'):
            self.set_flags(self.get_flags() | Gio.ApplicationFlags.NON_UNIQUE)
        if options.contains('version'):
            print(_("{program} version {version}").format(program=__program_name__, version=__version__))
        elif options.contains('copyright'):
            print(__copyright__)
            print(__license__)
        elif options.contains('list-actions'):
            self.register()
            for name in sorted(self.list_actions()):
                print(name)
        else:
            return -1
        return 0

    def do_startup(self):
        Gtk.Application.do_startup(self)
        logger.debug("Starting")
        asyncio.set_event_loop_policy(gbulb.gtk.GtkEventLoopPolicy())
        self.shared = shared.AppShared()
        self.shared.connect('notify::enable-fragile-accels', self.notify_enable_fragile_accels_cb)

        self.ampd = self.shared.ampd.sub_executor()
        self.shared.ampd_client.connect('client-connected', self.idle_output)

        self.notification = Gio.Notification.new(_("MPD status"))
        self.notification_task = None

        self.session_inhibit_cookie = None
        self.systemd_inhibit_fd = None
        self.shared.ampd_server_properties.connect('notify::state', self.set_inhibit)
        self.shared.connect('notify::restricted', self.set_inhibit)

        self.add_action(action.Action('new-window', self.new_window_cb))
        self.add_action(action.Action('close-window', self.close_window_cb))
        self.add_action(action.Action('new-module', self.new_module_cb, parameter_type=GLib.VariantType.new('s')))
        self.add_action(action.Action('close-module', self.close_module_cb))
        self.add_action(action.Action('help', self.help_cb))
        self.add_action(action.Action('about', self.about_cb))
        self.add_action(action.Action('notify', self.task_hold_app(self.action_notify_cb)))
        self.add_action(action.Action('quit', self.quit_cb))
        self.add_action(action.Action('edit-profiles', self.edit_profiles_cb))
        self.add_action(action.Action('manage-plugins', self.action_manage_plugins_cb))

        # self.add_action(action.Action('BAD', self.THIS_IS_BAD_cb))

        self.output_ids = []

        menubar = omenu.OrderedMenu(self.set_accels_for_action)
        self.set_menubar(menubar)
        menubar.insert_path('10#gampc', Gio.MENU_LINK_SUBMENU, _("_GAMPC"))
        menubar.insert_path('10#gampc/10#window/10', 'app.new-window', _("New window"), ['<Control>n'])
        menubar.insert_path('10#gampc/10#window/20', 'app.close-window', _("Close window"), ['<Control>w'])
        menubar.insert_path('10#gampc/10#window/30', 'win.toggle-fullscreen', _("Fullscreen window"), ['<Alt>f'])
        menubar.insert_path('10#gampc/10#window/40', 'win.volume-popup', _("Adjust volume"), ['<Alt>v'])
        menubar.insert_path('10#gampc/90#app/50', 'app.quit', _("Quit"), ['<Control>q'])
        menubar.insert_path('10#gampc/99#BAD/10', 'app.BAD', "BAD", ['<Control><Alt>z'])
        menubar.insert_path('20#edit', Gio.MENU_LINK_SUBMENU, _("_Edit"))
        menubar.insert_path('30#playback', Gio.MENU_LINK_SUBMENU, _("_Playback"))
        menubar.insert_path('40#server', Gio.MENU_LINK_SUBMENU, _("_Server"))
        menubar.insert_path('40#server/40#connection/30#profiles', Gio.MENU_LINK_SUBMENU, _("_Profiles"))
        menubar.insert_path('40#server/40#connection/30#profiles/20/10', 'app.edit-profiles', _("Edit profiles"))
        menubar.insert_path('50#modules', Gio.MENU_LINK_SUBMENU, _("_Modules"))
        menubar.insert_path('50#modules/10#modules', Gio.MENU_LINK_SECTION, _("Press <Ctrl> for a new instance"))
        menubar.insert_path('50#modules/20#current/10', 'app.close-module', _("Close module"), ['<Control><Shift>w'])
        menubar.insert_path('90#help', Gio.MENU_LINK_SUBMENU, _("_Help"))
        menubar.insert_path('90#help/10', 'app.help', _("Help"), hidden_when=None)
        menubar.insert_path('90#help/20', 'app.about', _("About"))
        self.setup_profile_menu()

        self.app_menu = omenu.OrderedMenu()
        self.set_app_menu(self.app_menu)
        self.app_menu.insert_path('10#window/10', 'app.new-window', _("New window"), ['<Control>n'])
        self.app_menu.insert_path('10#window/20', 'app.close-window', _("Close window"), ['<Control>w'])
        self.app_menu.insert_path('20#help/10', 'app.help', _("Help"), hidden_when=None)
        self.app_menu.insert_path('20#help/20', 'app.about', _("About"))
        self.app_menu.insert_path('30#plugins/20', 'app.manage-plugins', _("Manage plugins"))
        self.app_menu.insert_path('90#quit/10', 'app.quit', _("Quit"), ['<Control>q'])

        self.module_classes = {}
        self.modules = []

        self.shared.peas_extension_set.connect_after('extension-added', self.peas_extension_added_cb)
        self.shared.peas_extension_set.connect('extension-removed', self.peas_extension_removed_cb)
        self.shared.peas_engine.set_loaded_plugins(['core', 'current', 'playqueue', 'browser', 'search', 'savedsearch', 'playlist', 'tanda', 'command', 'log', 'tango-velours'])

        self.sigint_source = GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, signal.SIGINT, self.quit_cb)

    def do_command_line(self, command_line):
        options = command_line.get_options_dict().end().unpack()
        if 'debug' in options:
            logging.getLogger().setLevel(logging.DEBUG)
        if GLib.OPTION_REMAINING in options:
            for option in options[GLib.OPTION_REMAINING]:
                try:
                    success, name, target = Gio.Action.parse_detailed_name(option)
                except Exception as e:
                    logger.error(e)
                    continue
                if not self.has_action(name):
                    logger.error(_("Action '{name}' does not exist").format(name=name))
                else:
                    self.activate_action(name, target)
        else:
            self.activate()
        return 0

    def do_activate(self, *args):
        win = self.get_active_window()
        if win:
            win.present()
        else:
            self.new_window_cb(None, None)

    def do_shutdown(self):
        logger.debug("Shutting down")
        GLib.source_remove(self.sigint_source)
        for module in self.modules:
            module.destroy()
        self.shared.disconnect_by_func(self.notify_enable_fragile_accels_cb)
        self.shared.ampd_server_properties.disconnect_by_func(self.set_inhibit)
        self.shared.disconnect_by_func(self.set_inhibit)
        self.shared.peas_engine.set_loaded_plugins([])
        self.shared.close()
        for name in self.list_actions():
            self.remove_action(name)
        self.set_menubar()
        self.set_app_menu()
        Gtk.Application.do_shutdown(self)

    def task_hold_app(self, f):
        def g(*args, **kwargs):
            retval = f(*args, **kwargs)
            if isinstance(retval, asyncio.Future):
                self.hold()
                retval.add_done_callback(lambda future: self.release())
            return retval
        return g

    def notify_enable_fragile_accels_cb(self, shared, param):
        if shared.enable_fragile_accels:
            self.get_menubar().enable_fragile_accels()
        else:
            self.get_menubar().disable_fragile_accels()

    def peas_extension_added_cb(self, extension_set, info, extension):
        extension.shared = self.shared
        extension.activate()
        for module in extension.modules:
            self.register_module_class(module)
        provides = extension.provides.get('app', {})
        for item in provides.get('menubar_items', []):
            self.get_menubar().insert(item)
        for action_ in provides.get('actions', []):
            self.add_action(action_.create_action(self.task_hold_app, shared=self.shared) if isinstance(action_, action.ActionModel) else action_)

    def peas_extension_removed_cb(self, extension_set, info, extension):
        provides = extension.provides.get('app', {})
        for action_ in provides.get('actions', []):
            self.remove_action(action_.get_name())
        for item in reversed(provides.get('menubar_items', [])):
            self.get_menubar().remove_path(item.path)
        for module in extension.modules:
            self.unregister_module_class(module)
        extension.deactivate()
        del extension.shared

    def register_module_class(self, module_class):
        if module_class.name in self.module_classes:
            module_class.replace = self.module_classes[module_class.name]
            self.unregister_module_class(self.module_classes[module_class.name], restore=False)
        self.module_classes[module_class.name] = module_class
        self.get_menubar().insert_path('50#modules/10#modules/' + module_class.key,
                                       'app.new-module("{}")'.format(module_class.name),
                                       module_class.title,
                                       accels=['<Alt>' + module_class.key, '<Control><Alt>' + module_class.key])

    def unregister_module_class(self, module_class, restore=True):
        del self.module_classes[module_class.name]
        self.get_menubar().remove_path('50#modules/10#modules/' + module_class.key)
        if restore and module_class.replace:
            self.register_module_class(module_class.replace)
            module_class.replace = None

    def new_module_cb(self, action, parameter):
        self.start_module(parameter.unpack(), Gdk.Keymap.get_default().get_modifier_state() & Gdk.ModifierType.CONTROL_MASK, False)

    def close_module_cb(self, action, parameter):
        win = self.get_active_window()
        old_module = win.module
        if old_module is None:
            return
        for module in self.modules:
            if not module.win:
                win.change_module(module)
                break
        else:
            win.change_module(None)
        old_module.destroy()
        self.modules.remove(old_module)

    def new_window_cb(self, action, parameter):
        self.start_module('current', False, True)

    def close_window_cb(self, action, parameter):
        self.get_active_window().destroy()

    def start_module(self, name, new_instance, new_window):
        module = (not new_instance and self.pop_named_module(name)) or self.module_classes[name](self.shared)
        self.modules.append(module)
        win = None if new_window else module.win or self.get_active_window()
        if win is None:
            win = window.Window(self)
        if module.win is None:
            win.change_module(module)
        win.present()

    def pop_named_module(self, name):
        for module in self.modules:
            if module.name == name:
                self.modules.remove(module)
                return module
        return None

    def quit_cb(self, *args):
        logger.debug("Quit")
        for win in self.get_windows():
            win.destroy()
        return True

    def about_cb(self, *args):
        dialog = Gtk.AboutDialog(parent=self.get_active_window(), program_name=__program_name__, version=__version__, comments=__program_description__, copyright=__copyright__, license_type=Gtk.License.GPL_3_0, logo_icon_name='face-cool', website='http://math.univ-lyon1.fr/~begnac', website_label=_("Author's website"))
        dialog.run()
        dialog.destroy()

    def help_cb(self, *args):
        return
        builder = Gtk.Builder()
        win = Gtk.ShortcutsWindow(title='hhh')
        section = Gtk.ShortcutsSection(section_name='a', title='b')
        group = Gtk.ShortcutsGroup(title='yyy')
        shortcut = Gtk.ShortcutsShortcut(title='fwr')
        win.add_child(builder, section, None)
        section.add_child(builder, group, None)
        group.add_child(builder, shortcut, None)
        win.present()
        self.hold()

    def edit_profiles_cb(self, *args):
        struct = ssde.List(
            label=_("Profiles"),
            substruct=ssde.Dict(
                label=_("Profile"),
                substructs=[
                    ssde.Text(name='name', label=_("Profile name")),
                    ssde.Text(name='host', label=_("Host name")),
                    ssde.Integer(name='port', label=_("Port number"), default=6600),
                ]))
        value = struct.dialog(self.get_active_window(),
                              self.shared.config.server.profiles,
                              self.shared.config.server.access(self.shared.CONFIG_EDIT_DIALOG_SIZE, [500, 500]))
        if value:
            self.shared.config.server.profiles = value
            self.setup_profile_menu()

    def setup_profile_menu(self):
        menu = self.get_menubar().insert_path('40#server/40#connection/30#profiles/10#profiles', Gio.MENU_LINK_SECTION)
        menu.remove_all()
        for profile in self.shared.config.server.profiles:
            menu.append(profile['name'], 'app.server-profile("{name}")'.format_map(profile))

    @ampd.task
    async def action_notify_cb(self, *args):
        if self.notification_task:
            self.notification_task._close()
            self.withdraw_notification('status')
        self.notification_task = asyncio.Task.current_task()
        await self.ampd.idle(ampd.IDLE)
        print(self.shared.ampd_server_properties.current_song)
        if self.shared.ampd_server_properties.state == 'stop':
            icon_name = 'media-playback-stop-symbolic'
            body = 'Stopped'
        else:
            if self.shared.ampd_server_properties.state == 'play':
                icon_name = 'media-playback-start-symbolic'
            else:
                icon_name = 'media-playback-pause-symbolic'
            body = '{0} / {1}'.format(self.shared.ampd_server_properties.current_song.get('Artist', '???'), self.shared.ampd_server_properties.current_song.get('Title', '???'))
            if 'performer' in self.shared.ampd_server_properties.current_song:
                body += ' / ' + self.shared.ampd_server_properties.current_song['Performer']
        self.notification.set_body(body)
        self.notification.set_icon(Gio.Icon.new_for_string(icon_name))
        self.send_notification('status', self.notification)
        await asyncio.sleep(5)
        self.withdraw_notification('status')
        self.notification_task = None

    def set_inhibit(self, *args):
        if self.shared.ampd_server_properties.state == 'play':
            self.session_inhibit_cookie = self.session_inhibit_cookie or self.inhibit(None, Gtk.ApplicationInhibitFlags.SUSPEND | Gtk.ApplicationInhibitFlags.IDLE, __program_name__)
            bus = dbus.SystemBus()
            obj = bus.get_object('org.freedesktop.login1', '/org/freedesktop/login1')
            self.systemd_inhibit_fd = self.systemd_inhibit_fd or obj.Inhibit('handle-lid-switch', __program_name__, _("Playing in restricted mode"), 'block', dbus_interface='org.freedesktop.login1.Manager')
        else:
            self.session_inhibit_cookie = self.session_inhibit_cookie and self.uninhibit(self.session_inhibit_cookie)
            self.systemd_inhibit_fd = None

    @ampd.task
    async def idle_output(self, client):
        for output_id in self.output_ids:
            self.remove_action('output.' + output_id)
        self.output_ids = []
        menu = self.get_menubar().insert_path('40#server/30#output', Gio.MENU_LINK_SECTION, _("Outputs (<Ctrl> to toggle)"))
        menu.remove_all()
        outputs = await self.ampd.outputs()
        for output in outputs:
            output_id = output['outputid']
            self.output_ids.append(output_id)
            menu.insert_path(output_id, 'app.output.' + output_id, output['outputname'])
            self.add_action(action.Action('output.' + output_id, self.action_output_activate_cb, state=GLib.Variant.new_boolean(int(output['outputenabled']))))
        while True:
            await self.ampd.idle(ampd.OUTPUT)
            outputs = await self.ampd.outputs()
            for output in outputs:
                self.lookup_action('output.' + output['outputid']).set_state(GLib.Variant.new_boolean(int(output['outputenabled'])))

    @ampd.task
    async def action_output_activate_cb(self, action, parameter):
        output_id = action.get_name().split('.', 1)[1]
        if Gdk.Keymap.get_default().get_modifier_state() & Gdk.ModifierType.CONTROL_MASK:
            await self.ampd.toggleoutput(output_id)
        else:
            await self.ampd.enableoutput(output_id)
            await self.ampd.command_list(self.ampd.disableoutput(x) for x in self.output_ids if x is not output_id)

    def action_manage_plugins_cb(self, action, param):
        win = Gtk.Window()
        win.add(PeasGtk.PluginManager.new(self.shared.peas_engine))
        win.show_all()



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Graphical Asyncronous Music Player Client

A Music Player Daemon client written in Python/Gtk+3, using asynchronous communication.
"""

import gi
import os
import gettext


def setup_locale():
    gettext.install('gampc', os.path.join(os.path.dirname(__file__), 'locale'))


gi.require_version('Gtk', '3.0')
gi.require_version('Peas', '1.0')
gi.require_version('PeasGtk', '1.0')

setup_locale()


__program_name__ = "Graphical Asyncronous Music Player Client"
__program_description__ = _("A Music Player Daemon client written in Python/Gtk+3, using asynchronous communication")
__author__ = "Itaï BEN YAACOV"
__copyright__ = "© 2017 " + __author__
__license__ = _("""
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
""")
__version__ = '0.2.7'



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, Gtk, Gdk, GdkPixbuf

import xdg
import time
import asyncio
import ampd

from gampc import module
from gampc import plugin


class PixbufCache(dict):
    def __missing__(self, key):
        pixbuf = self.find_image(key)
        if pixbuf is not None:
            self[key] = pixbuf
        return pixbuf

    def find_image(self, key):
        for extension in ('.jpg', '.png', '.gif'):
            for name in xdg.BaseDirectory.load_data_paths('gampc', 'photos', key + extension):
                return GdkPixbuf.Pixbuf.new_from_file(name)

        for sep in (', ', ' y '):
            if sep in key:
                return self.find_images(key.split(sep))

        return None

    def find_images(self, names):
        pixbufs = [self[name] for name in names]
        if not all(pixbufs):
            return None
        width = height = 0
        for p in pixbufs:
            p.w = p.get_width()
            p.h = p.get_height()
            p.r = p.w / p.h
            height = max(height, 2 * p.h)
        for p in pixbufs:
            p.nw = height * p.r
            p.x = width
            width += p.nw
        pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, False, 8, width, height)
        for p in pixbufs:
            p.composite(pixbuf, p.x, 0, p.nw, height, p.x, 0, height / p.h, height / p.h, GdkPixbuf.InterpType.BILINEAR, 255)
        return pixbuf


class Current(module.Module):
    title = _("Current Song")
    name = 'current'
    key = '0'

    FIELD_NAMES = ['Artist', 'Performer', 'Title', 'Date', 'Genre', 'Composer']

    size = GObject.property(type=int)

    def __init__(self, *args):
        super().__init__(*args)
        self.signals['check-resize'] = self.window_check_resize_cb

        self.pixbufs = PixbufCache()
        self.images = []

        builder = self.shared.build_ui('plugins/current')
        self.welcome = builder.get_object('welcome')
        self.current = builder.get_object('current')
        self.shared.ampd_server_properties.bind_property('current-song', self.welcome, 'visible', GObject.BindingFlags.SYNC_CREATE, lambda x, y: not y)
        self.shared.ampd_server_properties.bind_property('current-song', self.current, 'visible', GObject.BindingFlags.SYNC_CREATE, lambda x, y: bool(y))
        self.shared.ampd_server_properties.connect('notify::current-song', self.notify_opacity_cb)
        self.running_fader = None

        self.width = 0
        self.css_provider = Gtk.CssProvider.new()
        self.get_style_context().add_provider(self.css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        self.connect('size-allocate', self.size_allocate_cb)
        self.bind_property('size', builder.get_object('image-icon'), 'pixel-size', GObject.BindingFlags(0), lambda x, y: y * 5)

        for field in self.FIELD_NAMES:
            label = builder.get_object('label-' + field.lower())
            label.field = field
            image = builder.get_object('image-' + field.lower())
            if image:
                image_label = builder.get_object('image-label-' + field.lower())
                if image_label:
                    label.bind_property('label', image_label, 'label', GObject.BindingFlags.SYNC_CREATE)
                    image.bind_property('visible', image_label, 'visible', GObject.BindingFlags.SYNC_CREATE)
                label.connect('notify::label', self.notify_label_cb, image)
                image.connect('size-allocate', self.image_size_allocate_cb)
                self.images.append(image)
            else:
                self.shared.ampd_server_properties.bind_property('current-song',
                                                                 label, 'visible',
                                                                 GObject.BindingFlags.SYNC_CREATE,
                                                                 lambda x, y, z: z in y,
                                                                 None, field)
            self.shared.ampd_server_properties.bind_property('current-song',
                                                             label, 'label',
                                                             GObject.BindingFlags.SYNC_CREATE,
                                                             lambda x, y, z: self.set_size() or y.get(z, ''),
                                                             None, field)

        self.add(builder.get_object('widget'))
        self.connect('map', self.__map_cb)
        self.connect('destroy', self.__destroy_cb)

    @staticmethod
    def __destroy_cb(self):
        if self.running_fader:
            self.running_fader.cancel()
            self.running_fader = None

    @staticmethod
    def __map_cb(self):
        self.width = 0
        self.set_size()

    def window_check_resize_cb(self, win):
        for image in self.images:
            image.clear()
            image.last_width = image.last_height = None

    def notify_label_cb(self, label, param, image):
        text = label.get_text()
        if not text or (label.field == 'Performer' and text == 'Instrumental'):
            label.get_parent().set_visible(False)
            return

        image.pixbuf = self.pixbufs[text]
        if image.pixbuf:
            image.last_width = image.last_height = None
            image.set_visible(True)
            label.set_visible(False)
        else:
            image.set_visible(False)
            label.set_visible(True)
        label.get_parent().set_visible(True)

    def notify_opacity_cb(self, *args):
        if self.running_fader:
            self.running_fader.cancel()
            self.running_fader = None
        if self.shared.ampd_server_properties.current_song:
            self.running_fader = self.fader()

    @ampd.task
    async def fader(self):
        START = 30
        DURATION = 3
        INTERVAL = 0.05

        if self.shared.dark:
            self.current.set_opacity(0)
            await asyncio.sleep(START)
            t0 = t1 = time.time()
            while t1 < t0 + DURATION:
                self.current.set_opacity((t1 - t0) / DURATION)
                await asyncio.sleep(INTERVAL)
                t1 = time.time()
        self.current.set_opacity(1)

    def do_button_press_event(self, event):
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            self.win.action_toggle_fullscreen_cb()

    @staticmethod
    def size_allocate_cb(self, allocation):
        self.width = allocation.width
        self.set_size()

    def set_size(self):
        scale = 50.0
        song = self.shared.ampd_server_properties.current_song
        if song:
            scale += 3 * max(len(song.get('Artist', '')) - 20, len(song.get('Title', '')) - 20, 0)
        self.size = self.width / scale
        css = b'* { font-size: ' + str(self.size).encode() + b'px; }'
        self.css_provider.load_from_data(css)

    def image_size_allocate_cb(self, image, allocation):
        if self.width == 0 or image.last_width == allocation.width and image.last_height == allocation.height:
            return
        image.last_width = allocation.width
        image.last_height = allocation.height
        ratio = image.pixbuf.get_height() / image.pixbuf.get_width()
        if allocation.width * ratio < allocation.height:
            allocation.height = allocation.width * ratio
        else:
            allocation.width = allocation.height / ratio
        pixbuf = image.pixbuf.scale_simple(allocation.width, allocation.height, GdkPixbuf.InterpType.BILINEAR)
        image.set_from_pixbuf(pixbuf)


class CurrentExtension(plugin.Extension):
    modules = [Current]



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, GLib, Gdk, Gtk, Pango
import re
import ast
import cairo


def format_time(time):
    time = int(time)
    hours = time // 3600
    minutes = (time // 60) % 60
    seconds = time % 60
    return '{:d}:{:02d}:{:02d}'.format(hours, minutes, seconds) if hours else '{:d}:{:02d}'.format(minutes, seconds)


def store_set_rows(store, rows, func):
    if not rows:
        store.clear()
        return
    i = store.get_iter_first()
    appending = (i is None)
    for row in rows:
        if appending:
            i = store.append()
        func(i, row)
        if not appending:
            i = store.iter_next(i)
            if i is None:
                appending = True
    if not appending:
        while store.remove(i):
            pass


def config_notify_cb(obj, param, config):
    config[param.name] = obj.get_property(param.name)


class Field(GObject.GObject):
    title = GObject.property(type=str)
    width = GObject.property(type=int)
    visible = GObject.property(type=bool, default=True)
    xalign = GObject.property(type=float, default=0.0)

    get_value = None

    def __init__(self, name, title=None, min_width=50, get_value=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.title = title
        self.width = self.min_width = min_width
        if get_value:
            self.get_value = get_value

    def get_renderer(self):
        return Gtk.CellRendererText(ellipsize=Pango.EllipsizeMode.END, xalign=self.xalign)

    def __repr__(self):
        return "Field '{title}'".format(title=self.title)


class FieldWithTable(Field):
    def __init__(self, name, title=None, table=None, min_width=50, **kwargs):
        super(FieldWithTable, self).__init__(name, title, min_width, **kwargs)
        self.table = table

    def get_value(self, record):
        for field, pattern, value in self.table:
            if not field:
                return value
            else:
                match = re.search(pattern, record.get(field, ''))
                if match:
                    return match.expand(value)


class FieldFamily(GObject.GObject):
    order = GObject.property()

    def __init__(self, config):
        super(FieldFamily, self).__init__()
        self.config = config
        self.old_order = self.config.access_base('order', [])
        self.order = self.config.order = []
        self.connect('notify', config_notify_cb, self.config)

        self.names = []
        self.basic_names = []
        self.derived_names = []
        self.fields = {}

    def close(self):
        self.disconnect_by_func(config_notify_cb)

    def register_field(self, field):
        if field.name in self.names:
            raise RuntimeError("Field '{name}' already registered".format(name=field.name))
        self.names.append(field.name)
        if field.get_value:
            self.derived_names.append(field.name)
        else:
            self.basic_names.append(field.name)
        self.fields[field.name] = field

        field_config = self.config.info[field.name]
        field.connect('notify', config_notify_cb, field_config)
        field.width = field_config.access('width', field.width)
        field.visible = field_config.access('visible', field.visible)
        if field.name in self.order:
            return
        if field.name not in self.old_order:
            self.order.append(field.name)
            return
        pos = self.old_order.index(field.name)
        for i, name in enumerate(self.order):
            if name not in self.old_order or self.old_order.index(name) > pos:
                self.order.insert(i, field.name)
                return
        else:
            self.order.append(field.name)

    def unregister_field(self, field):
        self.names.remove(field.name)
        if field.name in self.basic_names:
            self.basic_names.remove(field.name)
        else:
            self.derived_names.remove(field.name)
        del self.fields[field.name]
        field.disconnect_by_func(config_notify_cb)

    def record_set_fields(self, record):
        for name in self.derived_names:
            value = self.fields[name].get_value(record)
            if value is not None:
                record[name] = value

    def records_set_fields(self, records):
        for record in records:
            self.record_set_fields(record)


class FieldColumn(Gtk.TreeViewColumn):
    width_rw = GObject.property(type=int)

    def __init__(self, field, next_data_func):
        super(FieldColumn, self).__init__()
        self.field = field
        self.renderer = field.get_renderer()
        self.next_data_func = next_data_func
        field.bind_property('title', self, 'title', GObject.BindingFlags.SYNC_CREATE)
        field.bind_property('visible', self, 'visible', GObject.BindingFlags.SYNC_CREATE)
        self.set_min_width(field.min_width)
        self.connect('notify::width-rw', self.notify_width_rw_cb)
        field.bind_property('width', self, 'width-rw', GObject.BindingFlags.SYNC_CREATE)
        self.bind_property('width', field, 'width')
        self.set_resizable(True)
        self.set_reorderable(True)
        self.pack_start(self.renderer, True)
        self.set_cell_data_func(self.renderer, self.data_func)

    @staticmethod
    def notify_width_rw_cb(self, detail):
        if self.width_rw != self.get_width():
            self.set_fixed_width(self.width_rw)

    @staticmethod
    def data_func(self, renderer, store, i, arg):
        value = getattr(store.get_record(i) or Record(), self.field.name)
        renderer.set_property('text', str(value) if value is not None else None)
        self.next_data_func and self.next_data_func(self, renderer, store, i, arg)


class Record(GObject.Object):
    def __init__(self, data=None):
        super(Record, self).__init__()
        self.set_data(data or {})

    def set_data(self, data):
        GObject.Object.__setattr__(self, '_data', data)

    def get_data(self):
        return self._data

    def get_data_clean(self):
        return {key: value for key, value in self._data.items() if key[0] != '_'}

    def __getattr__(self, name):
        return self._data.get(name)

    def __setattr__(self, name, value):
        self._data[name] = value

    def __delattr__(self, name):
        self._data.pop(name, None)


class StoreBase(object):
    def __iter__(self):
        i = self.iter_children(None)
        while i:
            yield i, self.get_path(i), self.get_record(i)
            i = self.iter_next(i)

    def delete_refs(self, refs):
        for ref in refs:
            i = self.get_iter(ref.get_path())
            self.emit('record-delete', i)

    def get_record(self, i):
        return self.get_value(i, 0)

    def set_row(self, i, row):
        self.set_value(i, 0, Record(row))

    def __getattr__(self, name):
        return getattr(self.store, name)


class RecordStore(StoreBase, Gtk.ListStore):
    def __init__(self):
        super().__init__(Record)

    def set_rows(self, rows):
        store_set_rows(self, rows, self.set_row)


class RecordStoreFilter(StoreBase, Gtk.TreeModelFilter):
    filter_active = GObject.Property(type=bool, default=False)

    __gsignals__ = {
        'record-new': (GObject.SIGNAL_ACTION, None, (Gtk.TreeIter,)),
        'record-delete': (GObject.SIGNAL_ACTION, None, (Gtk.TreeIter,)),
    }

    def __init__(self):
        self.store = RecordStore()
        super().__init__(child_model=self.store)

    def remove(self, i):
        return self.store.remove(self.convert_iter_to_child_iter(i))

    def insert_after(self, i):
        if self.filter_active:
            raise Exception(_("Cannot add to a filtered list"))
        success, j = self.convert_child_iter_to_iter(self.store.insert_after(None if i is None else self.convert_iter_to_child_iter(i)))
        return j


class RecordStoreSort(StoreBase, Gtk.TreeModelSort):
    filter_active = GObject.Property(type=bool, default=False)

    __gsignals__ = {
        'record-delete': (GObject.SIGNAL_ACTION, None, (Gtk.TreeIter,)),
    }

    def __init__(self):
        self.store = RecordStoreFilter()
        super().__init__(self.store)
        self.bind_property('filter-active', self.store, 'filter-active')


class AutoScrollTreeView(Gtk.TreeView):
    def __init__(self, *args, **kwargs):
        self.key_just_pressed = None

        super().__init__(*args, **kwargs)

        self.connect('key-press-event', self.key_press_event_cb)
        self.connect('cursor-changed', self.cursor_changed_cb)
        self.connect('destroy', lambda self: self.key_just_pressed and GLib.source_remove(self.key_just_pressed))

    @staticmethod
    def key_press_event_cb(self, event):
        if event.type == Gdk.EventType.KEY_PRESS:
            if self.key_just_pressed:
                GLib.source_remove(self.key_just_pressed)
            self.key_just_pressed = GLib.timeout_add(200, self.key_press_event_timeout)

    def key_press_event_timeout(self):
        self.key_just_pressed = None

    @staticmethod
    def cursor_changed_cb(self):
        if not self.key_just_pressed:
            return
        path, col = self.get_cursor()
        if path:
            self.scroll_to_cell(path, None, True, 0.5, 0.5)


class RecordTreeViewBase(Gtk.TreeView):
    def __init__(self, model_class, fields, data_func=None, **kwargs):
        super().__init__(model=model_class(), search_column=0, enable_search=False, enable_grid_lines=Gtk.TreeViewGridLines.BOTH, **kwargs)

        self.fields = fields
        self.cols = {name: FieldColumn(self.fields.fields[name], data_func) for name in self.fields.order}
        for name in self.fields.order:
            self.append_column(self.cols[name])

        self.connect('destroy', self.destroy_cb)
        self.connect('columns-changed', self.columns_changed_cb)
        self.fields.connect('notify::order', self.fields_notify_order_cb)

    @staticmethod
    def destroy_cb(self):
        self.fields.disconnect_by_func(self.fields_notify_order_cb)
        self.disconnect_by_func(self.columns_changed_cb)
        del self.cols

    @staticmethod
    def columns_changed_cb(self):
        self.fields.handler_block_by_func(self.fields_notify_order_cb)
        self.fields.order = [self.get_column(i).field.name for i in range(self.get_n_columns())]
        self.fields.handler_unblock_by_func(self.fields_notify_order_cb)

    def fields_notify_order_cb(self, *args):
        self.handler_block_by_func(self.columns_changed_cb)
        last_col = None
        for name in self.fields.order:
            col = self.cols[name]
            self.move_column_after(col, last_col)
            last_col = col
        self.handler_unblock_by_func(self.columns_changed_cb)


class RecordTreeView(RecordTreeViewBase, AutoScrollTreeView):
    def __init__(self, fields, data_func, sortable):
        super().__init__(RecordStoreSort if sortable else RecordStoreFilter, fields, data_func, visible=True, vexpand=True, rubber_banding=True)
        self.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        self.sortable = sortable

        self.set_search_equal_func(lambda store, col, key, i: not any(isinstance(value, str) and key.lower() in value.lower() for value in store.get_record(i).get_data().values()))

        if self.sortable:
            store = self.get_model()
            for i, name in enumerate(self.fields.order):
                self.cols[name].set_sort_column_id(i)
                store.set_sort_func(i, self.sort_func, name)

        self.connect('drag-data-get', self.drag_data_get_cb)

    @staticmethod
    def sort_func(store, i, j, name):
        try:
            v1 = getattr(store.get_record(i), name)
            v2 = getattr(store.get_record(j), name)
            return 0 if v1 == v2 else -1 if v1 is None or (v2 is not None and v1 < v2) else 1
        except:
            return 0

    def get_selection_rows(self):
        store, paths = self.get_selection().get_selected_rows()
        return [store.get_record(store.get_iter(p)).get_data_clean() for p in paths], [Gtk.TreeRowReference.new(store, p) for p in paths]

    def clipboard_paste_cb(self, clipboard, raw, before):
        path, column = self.get_cursor()
        try:
            records = ast.literal_eval(raw)
        except:
            return
        if not (isinstance(records, list) and all(isinstance(record, dict) for record in records)):
            return
        self.paste_at(records, path, before)

    def paste_at(self, records, path, before):
        selection = self.get_selection()
        selection.unselect_all()
        store = self.get_model()
        i = store.get_iter(path) if path else None
        if i and before:
            i = store.iter_previous(i)
        cursor_set = False
        for record in records:
            j = store.insert_after(i)
            store.set_row(j, record)
            ref = Gtk.TreeRowReference.new(store, store.get_path(j))
            store.emit('record-new', j)
            i = store.get_iter(ref.get_path())
            if not cursor_set:
                cursor_set = True
                self.set_cursor(store.get_path(i))
            selection.select_iter(i)

    def do_drag_begin(self, context):
        drag_records, context.drag_refs = self.get_selection_rows()
        context.data = repr(drag_records).encode()
        if not drag_records:
            return
        icons = [self.create_row_drag_icon(ref.get_path()) for ref in context.drag_refs]
        width, height = icons[0].get_width(), icons[0].get_height()
        target = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height * len(context.drag_refs))
        cr = cairo.Context(target)
        cr.set_source_rgba(0, 0, 0, 1)
        cr.paint()
        y = 2
        for icon in icons:
            cr.set_source_surface(icon, 2, y)
            cr.paint()
            y += height
        icon.flush()
        Gtk.drag_set_icon_surface(context, target)

    @staticmethod
    def drag_data_get_cb(self, context, data, info, time):
        data.set(data.get_target(), 8, context.data)

    def do_drag_data_delete(self, context):
        self.get_model().delete_refs(context.drag_refs)
        context.drag_refs = []

    def do_drag_data_received(self, context, x, y, data, info, time):
        path, pos = self.get_dest_row_at_pos(x, y)
        records = ast.literal_eval(data.get_data().decode())
        self.paste_at(records, path, pos in [Gtk.TreeViewDropPosition.BEFORE, Gtk.TreeViewDropPosition.INTO_OR_BEFORE])

    def do_drag_end(self, context):
        del context.drag_refs

    def do_drag_motion(self, context, x, y, time):
        dest = self.get_dest_row_at_pos(x, y)
        if dest is None:
            return False
        self.set_drag_dest_row(*dest)
        if context.get_actions() & Gdk.DragAction.MOVE and not Gdk.Keymap.get_default().get_modifier_state() & Gdk.ModifierType.CONTROL_MASK:
            action = Gdk.DragAction.MOVE
        else:
            action = Gdk.DragAction.COPY
        Gdk.drag_status(context, action, time)
        return True


class RecordTreeViewFilter(RecordTreeViewBase):
    __gsignals__ = {
        'changed': (GObject.SIGNAL_RUN_LAST, None, ()),
    }

    def __init__(self, shared, filter_, fields, **kwargs):
        super().__init__(RecordStore, fields, self.data_func, can_focus=True, **kwargs)
        self.shared = shared
        store = self.get_model()
        self.filter_ = filter_
        store.set_value(store.append(), 0, filter_)
        self.get_selection().set_mode(Gtk.SelectionMode.NONE)
        for name, col in self.cols.items():
            col.renderer.set_property('editable', True)
            col.renderer.connect('editing-started', self.renderer_editing_started_cb, name)
        self.connect('button-press-event', self.button_press_event_cb)

    @staticmethod
    def button_press_event_cb(self, event):
        pos = self.get_path_at_pos(event.x, event.y)
        if not pos:
            return False
        path, col, cx, xy = pos
        self.set_cursor_on_cell(path, col, col.renderer, True)
        return True

    def renderer_editing_started_cb(self, renderer, editable, path, name):
        editable.connect('editing-done', self.editing_done_cb, name)
        self.shared.enable_fragile_accels = False
        self.handler_block_by_func(self.button_press_event_cb)

    def editing_done_cb(self, editable, name):
        self.handler_unblock_by_func(self.button_press_event_cb)
        self.shared.enable_fragile_accels = True
        if editable.get_property('editing-canceled'):
            return
        value = editable.get_text() or None
        if value != getattr(self.filter_, name):
            if value:
                setattr(self.filter_, name, value)
            else:
                delattr(self.filter_, name)
            self.emit('changed')

    green = Gdk.RGBA()
    green.parse('pale green')
    yellow = Gdk.RGBA()
    yellow.parse('yellow')

    @staticmethod
    def data_func(column, renderer, store, i, arg):
        renderer.set_property('background-rgba', RecordTreeViewFilter.green if renderer.get_property('text') is None else RecordTreeViewFilter.yellow)


class TreeViewFilter(Gtk.Box):
    active = GObject.Property(type=bool, default=False)

    def __init__(self, shared, treeview):
        super().__init__(visible=True, orientation=Gtk.Orientation.VERTICAL)

        self.filter_ = Record()
        self.filter_treeview = RecordTreeViewFilter(shared, self.filter_, treeview.fields)

        filter_scroller = Gtk.ScrolledWindow(visible=True)
        filter_scroller.add(self.filter_treeview)
        filter_scroller.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.NEVER)
        self.pack_start(filter_scroller, False, False, 0)

        self.scroller = Gtk.ScrolledWindow(visible=True)
        self.scroller.add(treeview)
        self.pack_start(self.scroller, True, True, 0)
        self.treeview = treeview
        treeview.bind_property('hadjustment', self.filter_treeview, 'hadjustment', GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL)
        self.bind_property('active', self.filter_treeview, 'visible')
        self.bind_property('active', treeview, 'headers-visible', GObject.BindingFlags.INVERT_BOOLEAN)
        self.bind_property('active', treeview.get_model(), 'filter-active')
        self.connect('notify::active', self.notify_active_cb)
        self.filter_treeview.connect('changed', lambda _: self.treeview.get_model().refilter())
        self.treeview.get_model().set_visible_func(self.visible_func)
        self.active = False

    @staticmethod
    def notify_active_cb(self, param):
        self.treeview.get_model().refilter()

    def visible_func(self, store, i, _):
        if not self.active:
            return True
        record = store.get_record(i)
        if record is None:
            return False
        for key, value in self.filter_.get_data().items():
            if re.search(value, getattr(record, key) or '', re.IGNORECASE) is None:
                return False
        return True



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk

import ampd

from gampc import plugin
from gampc.utils import action
from gampc.utils import omenu

import songlist


class Search(songlist.EditableFileSongList):
    title = _("Search")
    name = 'search'
    key = '3'

    duplicate_test_columns = ['Title', 'Artist', 'Performer', 'Date']

    sortable = True

    def __init__(self, *args, **kwargs):
        super(Search, self).__init__(*args, **kwargs)
        widget = self.get_child()
        self.remove(widget)

        self.entry = Gtk.SearchEntry(visible=True)
        self.entry.connect('activate', self.entry_activate_cb)
        self.entry.connect('focus-in-event', self.entry_focus_cb)
        self.entry.connect('focus-out-event', self.entry_focus_cb)
        self.field_choice = Gtk.ComboBoxText(visible=True)
        self.field_choice.append_text(_("any field"))
        for name in self.fields.names:
            self.field_choice.append_text(name)
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True)
        box.add(widget)
        box.add(self.entry)
        self.add(box)

        self.connect('map', self.action_search_cb)
        self.actions.add_action(action.Action('search', self.action_search_cb))

    def action_search_cb(self, *args):
        self.entry.grab_focus()

    @ampd.task
    async def client_connected_cb(self, client):
        while True:
            self.entry.activate()
            await self.ampd.idle(ampd.DATABASE)

    def action_reset_cb(self, action, parameter):
        super().action_reset_cb(action, parameter)
        self.field_choice.set_active(0)
        self.entry.activate()

    @ampd.task
    async def entry_activate_cb(self, entry):
        query = entry.get_text()
        if not query:
            return
        if query[0] == '!':
            query = query[1:]
            find = True
        else:
            find = False
        condition = sum((['any', s] if '=' not in s else s.split('=', 1) for s in self.parse(query)), [])
        if condition:
            songs = await (self.ampd.find if find else self.ampd.search)(*condition)
            self.set_songs(songs)

    def entry_focus_cb(self, entry, event):
        self.shared.enable_fragile_accels = not event.in_
        return False

    @staticmethod
    def parse(s):
        token = None
        for c in s:
            if token is None:
                if c.isspace():
                    continue
                else:
                    token = ''
                    quote = False
                    escape = False
            if escape:
                token += c
                escape = False
            elif c == '\\':
                escape = True
            elif c == '"':
                quote = not quote
            elif c.isspace() and not quote:
                yield token
                token = None
            else:
                token += c
        if token is not None:
            if quote:
                raise ValueError(_("Unbalanced quotes"))
            yield token


class SearchExtension(plugin.Extension):
    modules = [Search]

    def activate(self):
        self.provides['app'] = {}
        self.provides['app']['menubar_items'] = [omenu.Item('20#edit/90#search/10', 'mod.search', _("Search"), ['<Control><Shift>f'])]



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, Gio, Gtk, Gdk
import types
import asyncio

from . import data
from .utils import omenu
from .utils import config
from .utils.logger import logger


class Module(Gtk.Bin):
    title = None
    name = None
    key = None
    action_prefix = 'mod'
    use_provided = []
    replace = None

    status = GObject.property()
    full_title = GObject.property(type=str)

    def __init__(self, shared):
        super(Module, self).__init__(visible=True, full_title=self.title)
        self.signal_handlers = []

        self.bind_property('status', self, 'full-title', GObject.BindingFlags(0), lambda x, y: "{} [{}]".format(self.title, self.status) if self.status else self.title)

        self.connect('destroy', self.__destroy_cb)
        self.connect('map', self.__map_cb)
        self.connect('unmap', self.__unmap_cb)
        self.shared = shared
        self.config = config.ConfigDict.load(self.name, shared.config)
        self.ampd = self.shared.ampd_client.executor.sub_executor()
        self.signal_handler_connect(self.shared.ampd_client, 'client-connected', self.client_connected_cb)
        if self.shared.ampd_client.is_connected:
            self.client_connected_cb(self.shared.ampd_client)
        self.actions = Gio.SimpleActionGroup()
        self.signals = {}
        self.win = None
        self.provided = shared.collect_plugin_provides([self.name] + self.use_provided)
        for model in self.provided.get('actions', []):
            self.actions.add_action(model.create_action(lambda f: types.MethodType(f, self)))

    def __del__(self):
        logger.debug('Deleting {}'.format(self))

    @staticmethod
    def __destroy_cb(self):
        for target, handler in self.signal_handlers:
            target.disconnect(handler)
        del self.signal_handlers
        asyncio.ensure_future(self.ampd.close())
        del self.signals
        del self.actions

    def signal_handler_connect(self, target, *args):
        handler = target.connect(*args)
        self.signal_handlers.append((target, handler))

    def setup_context_menu(self, items, widget):
        if not items:
            return
        context_menu = omenu.OrderedMenu()
        for item in items:
            context_menu.insert(item)
        gtk_context_menu = Gtk.Menu.new_from_model(context_menu)
        gtk_context_menu.insert_action_group(self.action_prefix, self.actions)
        widget.connect('button-press-event', self.context_menu_button_press_event_cb, gtk_context_menu)

    @staticmethod
    def context_menu_button_press_event_cb(widget, event, context_menu):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 3:
            context_menu.popup(None, None, None, None, event.button, event.time)
            return True
        return False

    @staticmethod
    def client_connected_cb(client):
        pass

    @staticmethod
    def __map_cb(self):
        self.win = self.get_toplevel()
        self.win.insert_action_group(self.action_prefix, self.actions)
        for name, cb in self.signals.items():
            self.win.connect(name, cb)
        self.connect('notify::title-extra', self.win.update_title)

    @staticmethod
    def __unmap_cb(self):
        self.disconnect_by_func(self.win.update_title)
        if self.win.get_action_group(self.action_prefix) == self.actions:
            self.win.insert_action_group(self.action_prefix, None)
        for cb in self.signals.values():
            self.win.disconnect_by_func(cb)
        self.win = None


class PanedModule(Module):
    PANE_SEPARATOR_CONFIG = 'pane_separator'
    PANE_SEPARATOR_DEFAULT = 100

    def __init__(self, shared):
        super(PanedModule, self).__init__(shared)

        self.left_treeview = Gtk.TreeView(visible=True)
        self.scrolled_left_treeview = Gtk.ScrolledWindow(visible=True)
        self.scrolled_left_treeview.add(self.left_treeview)
        self.left_treeview.get_selection().connect('changed', self.left_treeview_selection_changed_cb)
        self.left_treeview.set_search_equal_func(lambda store, col, key, i: key.lower() not in store.get_value(i, col).lower())

        self.paned = Gtk.Paned(orientation=Gtk.Orientation.HORIZONTAL, position=self.config.access(self.PANE_SEPARATOR_CONFIG, self.PANE_SEPARATOR_DEFAULT), visible=True)
        self.paned.connect('notify::position', self.paned_notify_position_cb)
        self.paned.add1(self.scrolled_left_treeview)
        super(PanedModule, self).add(self.paned)

        self.setup_context_menu(self.provided.get('left_context_menu_items', []), self.left_treeview)

        self.starting = True
        self.connect('map', self.__map_cb)

    @staticmethod
    def __map_cb(self):
        if self.starting:
            self.left_treeview.grab_focus()
            self.starting = False

    def paned_notify_position_cb(self, *args):
        self.config[self.PANE_SEPARATOR_CONFIG] = self.paned.get_position()

    def left_store_set_rows(self, rows):
        data.store_set_rows(self.left_store, rows, lambda i, name: self.left_store.set_value(i, 0, name))

    def add(self, child):
        self.paned.add2(child)

    def remove(self, child):
        self.paned.remove2(child)



# coding: utf-8
#
# Graphical Asynchronous Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import apsw


class Cursor(object):
    def __init__(self, cursor, connection):
        self._cursor = cursor
        self._connection = connection

    def execute(self, *args, **kwargs):
        try:
            return self._cursor.execute(*args, **kwargs)
        except apsw.ReadOnlyError:
            self._connection._set_connection()
            raise

    def executemany(self, *args, **kwargs):
        try:
            return self._cursor.executemany(*args, **kwargs)
        except apsw.ReadOnlyError:
            self._connection._set_connection()
            raise


class Connection(object):
    def __init__(self, *args, **kwargs):
        self._init_args = args
        self._init_kwargs = kwargs
        self._set_connection()

    def _set_connection(self):
        self._connection = apsw.Connection(*self._init_args, **self._init_kwargs)

    def cursor(self):
        return Cursor(self._connection.cursor(), self)

    def __getattr__(self, name):
        return getattr(self._connection, name)

    def __enter__(self):
        return self._connection.__enter__()

    def __exit__(self, *args):
        return self._connection.__exit__(*args)
